from csle_common.dao.emulation_config.node_beats_config import NodeBeatsConfig
from csle_common.dao.emulation_config.beats_config import BeatsConfig
from csle_common.dao.emulation_config.client_managers_info import ClientManagersInfo
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.dao.emulation_config.cluster_node import ClusterNode
from csle_common.dao.emulation_config.cluster_config import ClusterConfig
from csle_common.dao.emulation_config.config import Config
from csle_common.dao.emulation_config.credential import Credential
from csle_common.dao.emulation_config.default_network_firewall_config import DefaultNetworkFirewallConfig
from csle_common.dao.emulation_config.node_firewall_config import NodeFirewallConfig
from csle_common.dao.emulation_config.docker_stats_manager_config import DockerStatsManagerConfig
from csle_common.dao.emulation_config.docker_stats_managers_info import DockerStatsManagersInfo
from csle_common.dao.emulation_config.node_network_config import NodeNetworkConfig
from csle_common.dao.emulation_config.host_manager_config import HostManagerConfig
from csle_common.dao.emulation_config.snort_ids_manager_config import SnortIDSManagerConfig
from csle_common.dao.emulation_config.node_resources_config import NodeResourcesConfig
from csle_common.dao.emulation_config.resources_config import ResourcesConfig
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.dao.emulation_config.node_services_config import NodeServicesConfig
from csle_common.dao.emulation_config.services_config import ServicesConfig
from csle_common.dao.emulation_config.users_config import UsersConfig
from csle_common.dao.emulation_config.traffic_config import TrafficConfig
from csle_common.dao.emulation_config.vulnerabilities_config import VulnerabilitiesConfig
from csle_common.dao.emulation_config.client_population_config import ClientPopulationConfig
from csle_common.dao.emulation_config.node_traffic_config import NodeTrafficConfig
from csle_common.dao.emulation_config.containers_config import ContainersConfig
from csle_common.dao.emulation_config.user import User
from csle_common.dao.emulation_config.node_users_config import NodeUsersConfig
from csle_common.dao.emulation_config.node_vulnerability_config import NodeVulnerabilityConfig
from csle_common.dao.emulation_config.flag import Flag
from csle_common.dao.emulation_config.node_flags_config import NodeFlagsConfig
from csle_common.dao.emulation_config.flags_config import FlagsConfig
from csle_common.dao.emulation_config.topology_config import TopologyConfig
from csle_common.dao.emulation_config.elk_config import ElkConfig
from csle_common.dao.emulation_config.elk_managers_info import ELKManagersInfo
from csle_common.dao.emulation_config.snort_managers_info import SnortIdsManagersInfo
from csle_common.dao.emulation_config.ossec_managers_info import OSSECIDSManagersInfo
from csle_common.dao.emulation_config.ryu_managers_info import RyuManagersInfo
from csle_common.dao.emulation_config.host_managers_info import HostManagersInfo
from csle_common.dao.emulation_config.kafka_managers_info import KafkaManagersInfo
from csle_common.dao.emulation_config.ossec_ids_manager_config import OSSECIDSManagerConfig
from csle_common.dao.emulation_config.sdn_controller_config import SDNControllerConfig
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.ovs_config import OVSConfig
from csle_common.dao.emulation_config.ovs_switch_config import OvsSwitchConfig
from csle_common.dao.emulation_config.kafka_config import KafkaConfig
from csle_common.dao.emulation_config.kafka_topic import KafkaTopic
from csle_common.dao.emulation_config.network_service import NetworkService
from csle_common.dao.emulation_observation.common.emulation_connection_observation_state import \
    EmulationConnectionObservationState
from csle_common.dao.emulation_config.connection_setup_dto import ConnectionSetupDTO
from csle_collector.client_manager.dao.client import Client
from csle_collector.client_manager.dao.constant_arrival_config import ConstantArrivalConfig
from csle_collector.client_manager.dao.workflow_markov_chain import WorkflowMarkovChain
from csle_collector.client_manager.dao.workflow_service import WorkflowService
from csle_collector.client_manager.dao.workflows_config import WorkflowsConfig


class TestEmulationConfigDaoSuite:
    """
    Test suite for datasets data access objects (DAOs)
    """

    def test_node_beats_config(self, example_node_beats_config: NodeBeatsConfig) -> None:
        """
        Tests creation and dict conversion of the NodeBeatsConfig DAO

        :param example_node_beats_config: an example NodeBeatsConfig
        :return: None
        """
        assert isinstance(example_node_beats_config.to_dict(), dict)
        assert isinstance(NodeBeatsConfig.from_dict(example_node_beats_config.to_dict()), NodeBeatsConfig)
        assert NodeBeatsConfig.from_dict(example_node_beats_config.to_dict()).to_dict() == \
               example_node_beats_config.to_dict()
        assert NodeBeatsConfig.from_dict(example_node_beats_config.to_dict()) == example_node_beats_config

    def test_beats_config(self, example_beats_config: BeatsConfig) -> None:
        """
        Tests creation and dict conversion of the BeatsConfig DAO

        :param example_beats_config: an example BeatsConfig
        :return: None
        """
        assert isinstance(example_beats_config.to_dict(), dict)
        assert isinstance(BeatsConfig.from_dict(example_beats_config.to_dict()), BeatsConfig)
        assert BeatsConfig.from_dict(example_beats_config.to_dict()).to_dict() == example_beats_config.to_dict()
        assert BeatsConfig.from_dict(example_beats_config.to_dict()) == example_beats_config

    def test_client_managers_info(self, example_client_managers_info: ClientManagersInfo) -> None:
        """
        Tests creation and dict conversion of the client_managers_info_dto

        :param example_client_managers_info: an example ClientManagersInfo
        :return: None
        """
        assert isinstance(example_client_managers_info.to_dict(), dict)
        assert isinstance(ClientManagersInfo.from_dict(example_client_managers_info.to_dict()), ClientManagersInfo)
        assert ClientManagersInfo.from_dict(example_client_managers_info.to_dict()).to_dict() == \
               example_client_managers_info.to_dict()
        assert ClientManagersInfo.from_dict(example_client_managers_info.to_dict()) == example_client_managers_info

    def test_container_network(self, example_container_network: ContainerNetwork) -> None:
        """
        Tests creation and dict conversion of the ContainerNetwork DTO

        :param example_container_network: an example ContainerNetwork
        :return: None
        """
        assert isinstance(example_container_network.to_dict(), dict)
        assert isinstance(ContainerNetwork.from_dict(example_container_network.to_dict()), ContainerNetwork)
        assert ContainerNetwork.from_dict(example_container_network.to_dict()).to_dict() == \
               example_container_network.to_dict()
        assert ContainerNetwork.from_dict(example_container_network.to_dict()) == \
               example_container_network

    def test_cluster_node(self, example_cluster_node: ClusterNode) -> None:
        """
        Tests creation and dict conversion of the ClusterNode DTO

        :param example_cluster_node: an example ClusterNode
        :return: None
        """
        assert isinstance(example_cluster_node.to_dict(), dict)
        assert isinstance(ClusterNode.from_dict(example_cluster_node.to_dict()), ClusterNode)
        assert ClusterNode.from_dict(example_cluster_node.to_dict()).to_dict() == example_cluster_node.to_dict()
        assert ClusterNode.from_dict(example_cluster_node.to_dict()) == example_cluster_node

    def test_cluster_config(self, example_cluster_config: ClusterConfig) -> None:
        """
        Tests creation and dict conversion of the ClusterConfig DTO

        :param example_cluster_config: an example ClusterConfig
        :return: None
        """
        assert isinstance(example_cluster_config.to_dict(), dict)
        assert isinstance(ClusterConfig.from_dict(example_cluster_config.to_dict()), ClusterConfig)
        assert ClusterConfig.from_dict(example_cluster_config.to_dict()).to_dict() == example_cluster_config.to_dict()
        assert ClusterConfig.from_dict(example_cluster_config.to_dict()) == example_cluster_config

    def test_config(self, example_config: Config) -> None:
        """
        Tests creation and dict conversion of the Config DTO

        :param example_config: an example Config
        :return: None
        """
        assert isinstance(example_config.to_dict(), dict)
        assert isinstance(Config.from_dict(example_config.to_dict()), Config)
        assert Config.from_dict(example_config.to_dict()).to_dict() == example_config.to_dict()
        assert Config.from_dict(example_config.to_dict()) == example_config

    def test_credential(self, example_credential: Credential) -> None:
        """
        Tests creation and dict conversion of the Credential DTO

        :param example_credential: an example Credential
        :return: None
        """
        assert isinstance(example_credential.to_dict(), dict)
        assert isinstance(Credential.from_dict(example_credential.to_dict()), Credential)
        assert Credential.from_dict(example_credential.to_dict()).to_dict() == example_credential.to_dict()
        assert Credential.from_dict(example_credential.to_dict()) == example_credential

    def test_default_network_firewall_config(
            self, example_default_network_firewall_config: DefaultNetworkFirewallConfig) -> None:
        """
        Tests creation and dict conversion of the DefaultNetworkFirewallConfig DTO

        :param example_default_network_firewall_config: an example DefaultNetworkFirewallConfig
        :return: None
        """
        assert isinstance(example_default_network_firewall_config.to_dict(), dict)
        assert isinstance(DefaultNetworkFirewallConfig.from_dict(example_default_network_firewall_config.to_dict()),
                          DefaultNetworkFirewallConfig)
        assert DefaultNetworkFirewallConfig.from_dict(example_default_network_firewall_config.to_dict()).to_dict() == \
               example_default_network_firewall_config.to_dict()
        assert DefaultNetworkFirewallConfig.from_dict(example_default_network_firewall_config.to_dict()) == \
               example_default_network_firewall_config

    def test_docker_stats_manager_config(self, example_docker_stats_manager_config: DockerStatsManagerConfig) -> None:
        """
        Tests creation and dict conversion of the DockerStatsManagerConfig DTO

        :param example_docker_stats_manager_config: an example DockerStatsManagerConfig
        :return: None
        """
        assert isinstance(example_docker_stats_manager_config.to_dict(), dict)
        assert isinstance(DockerStatsManagerConfig.from_dict(example_docker_stats_manager_config.to_dict()),
                          DockerStatsManagerConfig)
        assert DockerStatsManagerConfig.from_dict(example_docker_stats_manager_config.to_dict()).to_dict() == \
               example_docker_stats_manager_config.to_dict()
        assert DockerStatsManagerConfig.from_dict(example_docker_stats_manager_config.to_dict()) == \
               example_docker_stats_manager_config

    def test_docker_stats_managers_info(self, example_docker_stats_managers_info: DockerStatsManagersInfo) -> None:
        """
        Tests creation and dict conversion of the DockerStatsManagersInfo DTO

        :param example_docker_stats_managers_info: an example DockerStatsManagersInfo
        :return: None
        """
        assert isinstance(example_docker_stats_managers_info.to_dict(), dict)
        assert isinstance(DockerStatsManagersInfo.from_dict(example_docker_stats_managers_info.to_dict()),
                          DockerStatsManagersInfo)
        assert DockerStatsManagersInfo.from_dict(example_docker_stats_managers_info.to_dict()).to_dict() == \
               example_docker_stats_managers_info.to_dict()
        assert DockerStatsManagersInfo.from_dict(example_docker_stats_managers_info.to_dict()) == \
               example_docker_stats_managers_info

    def test_node_network_config(self, example_node_network_config: NodeNetworkConfig) -> None:
        """
        Tests creation and dict conversion of the NodeNetworkConfig DTO

        :param example_node_network_config: an example NodeNetworkConfig
        :return: None
        """
        assert isinstance(example_node_network_config.to_dict(), dict)
        assert isinstance(NodeNetworkConfig.from_dict(example_node_network_config.to_dict()), NodeNetworkConfig)
        assert NodeNetworkConfig.from_dict(example_node_network_config.to_dict()).to_dict() == \
               example_node_network_config.to_dict()
        assert NodeNetworkConfig.from_dict(example_node_network_config.to_dict()) == example_node_network_config

    def test_node_resources_config(self, example_node_resources_config: NodeResourcesConfig) -> None:
        """
        Tests creation and dict conversion of the NodeResourcesConfig DTO

        :param example_node_resources_config: an example NodeResourcesConfig
        :return: None
        """
        assert isinstance(example_node_resources_config.to_dict(), dict)
        assert isinstance(NodeResourcesConfig.from_dict(example_node_resources_config.to_dict()), NodeResourcesConfig)
        assert NodeResourcesConfig.from_dict(example_node_resources_config.to_dict()).to_dict() == \
               example_node_resources_config.to_dict()
        assert NodeResourcesConfig.from_dict(example_node_resources_config.to_dict()) == example_node_resources_config

    def test_node_container_config(self, example_node_container_config: NodeContainerConfig) -> None:
        """
        Tests creation and dict conversion of the NodeContainerConfig DTO

        :param example_node_container_config: an example NodeContainerConfig
        :return: None
        """
        assert isinstance(example_node_container_config.to_dict(), dict)
        assert isinstance(NodeContainerConfig.from_dict(example_node_container_config.to_dict()), NodeContainerConfig)
        assert NodeContainerConfig.from_dict(example_node_container_config.to_dict()).to_dict() == \
               example_node_container_config.to_dict()
        assert NodeContainerConfig.from_dict(example_node_container_config.to_dict()) == example_node_container_config

    def test_node_firewall_config(self, example_node_firewall_config: NodeFirewallConfig) -> None:
        """
        Tests creation and dict conversion of the NodeFirewallConfig DTO

        :param example_node_firewall_config: an example NodeFirewallConfig
        :return: None
        """
        assert isinstance(example_node_firewall_config.to_dict(), dict)
        assert isinstance(NodeFirewallConfig.from_dict(example_node_firewall_config.to_dict()), NodeFirewallConfig)
        assert NodeFirewallConfig.from_dict(example_node_firewall_config.to_dict()).to_dict() == \
               example_node_firewall_config.to_dict()
        assert NodeFirewallConfig.from_dict(example_node_firewall_config.to_dict()) == example_node_firewall_config

    def test_elk_config(self, example_elk_config: ElkConfig) -> None:
        """
        Tests creation and dict conversion of the ElkConfig DTO

        :param example_elk_config: an example ElkConfig
        :return: None
        """
        assert isinstance(example_elk_config.to_dict(), dict)
        assert isinstance(ElkConfig.from_dict(example_elk_config.to_dict()), ElkConfig)
        assert ElkConfig.from_dict(example_elk_config.to_dict()).to_dict() == example_elk_config.to_dict()
        assert ElkConfig.from_dict(example_elk_config.to_dict()) == example_elk_config

    def test_elk_managers_info(self, example_elk_managers_info: ELKManagersInfo) -> None:
        """
        Tests creation and dict conversion of the ElkManagersInfo DTO

        :param example_elk_managers_info: an example ELKManagersInfo
        :return: None
        """
        assert isinstance(example_elk_managers_info.to_dict(), dict)
        assert isinstance(ELKManagersInfo.from_dict(example_elk_managers_info.to_dict()),
                          ELKManagersInfo)
        assert ELKManagersInfo.from_dict(example_elk_managers_info.to_dict()).to_dict() == \
               example_elk_managers_info.to_dict()
        assert ELKManagersInfo.from_dict(example_elk_managers_info.to_dict()) == example_elk_managers_info

    def test_snort_managers_info(self, example_snort_ids_managers_info: SnortIdsManagersInfo) -> None:
        """
        Tests creation and dict conversion of the SnortManagersInfo DTO

        :param example_snort_ids_managers_info: an example SnortIdsManagersInfo
        :return: None
        """
        assert isinstance(example_snort_ids_managers_info.to_dict(), dict)
        assert isinstance(SnortIdsManagersInfo.from_dict(example_snort_ids_managers_info.to_dict()),
                          SnortIdsManagersInfo)
        assert SnortIdsManagersInfo.from_dict(example_snort_ids_managers_info.to_dict()).to_dict() == \
               example_snort_ids_managers_info.to_dict()
        assert SnortIdsManagersInfo.from_dict(example_snort_ids_managers_info.to_dict()) == \
               example_snort_ids_managers_info

    def test_ossec_managers_info(self, example_ossec_ids_managers_info: OSSECIDSManagersInfo) -> None:
        """
        Tests creation and dict conversion of the OSSECIDSManagersInfo DTO

        :param example_ossec_ids_managers_info: an example OSSECIDSManagersInfo
        :return: None
        """
        assert isinstance(example_ossec_ids_managers_info.to_dict(), dict)
        assert isinstance(OSSECIDSManagersInfo.from_dict(example_ossec_ids_managers_info.to_dict()),
                          OSSECIDSManagersInfo)
        assert OSSECIDSManagersInfo.from_dict(example_ossec_ids_managers_info.to_dict()).to_dict() == \
               example_ossec_ids_managers_info.to_dict()
        assert OSSECIDSManagersInfo.from_dict(example_ossec_ids_managers_info.to_dict()) == \
               example_ossec_ids_managers_info

    def test_ryu_managers_info(self, example_ryu_managers_info: RyuManagersInfo) -> None:
        """
        Tests creation and dict conversion of the RyuManagersInfo DTO

        :param example_ryu_managers_info: an example RyuManagersInfo
        :return: None
        """
        assert isinstance(example_ryu_managers_info.to_dict(), dict)
        assert isinstance(RyuManagersInfo.from_dict(example_ryu_managers_info.to_dict()),
                          RyuManagersInfo)
        assert RyuManagersInfo.from_dict(example_ryu_managers_info.to_dict()).to_dict() == \
               example_ryu_managers_info.to_dict()
        assert RyuManagersInfo.from_dict(example_ryu_managers_info.to_dict()) == example_ryu_managers_info

    def test_host_managers_info(self, example_host_managers_info: HostManagersInfo) -> None:
        """
        Tests creation and dict conversion of the HostManagersInfo DTO

        :param example_host_managers_info: an example HostManagersInfo
        :return: None
        """
        assert isinstance(example_host_managers_info.to_dict(), dict)
        assert isinstance(HostManagersInfo.from_dict(example_host_managers_info.to_dict()),
                          HostManagersInfo)
        assert HostManagersInfo.from_dict(example_host_managers_info.to_dict()).to_dict() == \
               example_host_managers_info.to_dict()
        assert HostManagersInfo.from_dict(example_host_managers_info.to_dict()) == example_host_managers_info

    def test_kafka_managers_info(self, example_kafka_managers_info: KafkaManagersInfo) -> None:
        """
        Tests creation and dict conversion of the KafkaManagersInfo DTO

        :param example_kafka_managers_info: an example KafkaManagersInfo
        :return: None
        """
        assert isinstance(example_kafka_managers_info.to_dict(), dict)
        assert isinstance(KafkaManagersInfo.from_dict(example_kafka_managers_info.to_dict()),
                          KafkaManagersInfo)
        assert KafkaManagersInfo.from_dict(example_kafka_managers_info.to_dict()).to_dict() == \
               example_kafka_managers_info.to_dict()
        assert KafkaManagersInfo.from_dict(example_kafka_managers_info.to_dict()) == example_kafka_managers_info

    def test_kafka_topic(self, example_kafka_topic: KafkaTopic) -> None:
        """
        Tests creation and dict conversion of the KafkaTopic DTO

        :param example_kafka_topic: an example KafkaTopic
        :return: None
        """
        assert isinstance(example_kafka_topic.to_dict(), dict)
        assert isinstance(KafkaTopic.from_dict(example_kafka_topic.to_dict()), KafkaTopic)
        assert KafkaTopic.from_dict(example_kafka_topic.to_dict()).to_dict() == example_kafka_topic.to_dict()
        assert KafkaTopic.from_dict(example_kafka_topic.to_dict()) == example_kafka_topic

    def test_kafka_config(self, example_kafka_config: KafkaConfig) -> None:
        """
        Tests creation and dict conversion of the KafkaConfig DTO

        :param example_kafka_config: an example KafkaConfig
        :return: None
        """
        assert isinstance(example_kafka_config.to_dict(), dict)
        assert isinstance(KafkaConfig.from_dict(example_kafka_config.to_dict()), KafkaConfig)
        assert KafkaConfig.from_dict(example_kafka_config.to_dict()).to_dict() == example_kafka_config.to_dict()
        assert KafkaConfig.from_dict(example_kafka_config.to_dict()) == example_kafka_config

    def test_network_service(self, example_network_service: NetworkService) -> None:
        """
        Tests creation and dict conversion of the NetworkService DTO

        :param example_network_service: an example NetworkService
        :return: None
        """
        assert isinstance(example_network_service.to_dict(), dict)
        assert isinstance(NetworkService.from_dict(example_network_service.to_dict()), NetworkService)
        assert NetworkService.from_dict(example_network_service.to_dict()).to_dict() == \
               example_network_service.to_dict()
        assert NetworkService.from_dict(example_network_service.to_dict()) == example_network_service

    def test_flag(self, example_flag: Flag) -> None:
        """
        Tests creation and dict conversion of the Flag DTO

        :param example_flag: an example Flag
        :return: None
        """
        assert isinstance(example_flag.to_dict(), dict)
        assert isinstance(Flag.from_dict(example_flag.to_dict()), Flag)
        assert Flag.from_dict(example_flag.to_dict()).to_dict() == example_flag.to_dict()
        assert Flag.from_dict(example_flag.to_dict()) == example_flag

    def test_node_flags_config(self, example_node_flags_config: NodeFlagsConfig) -> None:
        """
        Tests creation and dict conversion of the NodeFlagsConfig DTO

        :param example_node_flags_config: an example NodeFlagsConfig
        :return: None
        """
        assert isinstance(example_node_flags_config.to_dict(), dict)
        assert isinstance(NodeFlagsConfig.from_dict(example_node_flags_config.to_dict()), NodeFlagsConfig)
        assert NodeFlagsConfig.from_dict(example_node_flags_config.to_dict()).to_dict() == \
               example_node_flags_config.to_dict()
        assert NodeFlagsConfig.from_dict(example_node_flags_config.to_dict()) == example_node_flags_config

    def test_node_services_config(self, example_node_services_config: NodeServicesConfig) -> None:
        """
        Tests creation and dict conversion of the NodeServicesConfig DTO

        :param example_node_services_config: an example NodeServicesConfig
        :return: None
        """
        assert isinstance(example_node_services_config.to_dict(), dict)
        assert isinstance(NodeServicesConfig.from_dict(example_node_services_config.to_dict()), NodeServicesConfig)
        assert NodeServicesConfig.from_dict(example_node_services_config.to_dict()).to_dict() == \
               example_node_services_config.to_dict()
        assert NodeServicesConfig.from_dict(example_node_services_config.to_dict()) == example_node_services_config

    def test_node_traffic_config(self, example_node_traffic_config: NodeTrafficConfig) -> None:
        """
        Tests creation and dict conversion of the NodeTrafficConfig DTO

        :param example_node_traffic_config: an example NodeTrafficConfig
        :return: None
        """
        assert isinstance(example_node_traffic_config.to_dict(), dict)
        assert isinstance(NodeTrafficConfig.from_dict(example_node_traffic_config.to_dict()), NodeTrafficConfig)
        assert NodeTrafficConfig.from_dict(example_node_traffic_config.to_dict()).to_dict() == \
               example_node_traffic_config.to_dict()
        assert NodeTrafficConfig.from_dict(example_node_traffic_config.to_dict()) == example_node_traffic_config

    def test_user(self, example_user: User) -> None:
        """
        Tests creation and dict conversion of the User DTO

        :param example_user: an example User
        :return: None
        """
        assert isinstance(example_user.to_dict(), dict)
        assert isinstance(User.from_dict(example_user.to_dict()), User)
        assert User.from_dict(example_user.to_dict()).to_dict() == example_user.to_dict()
        assert User.from_dict(example_user.to_dict()) == example_user

    def test_node_users_config(self, example_node_users_config: NodeUsersConfig) -> None:
        """
        Tests creation and dict conversion of the NodeUsersConfig DTO

        :param example_node_users_config: an example NodeUsersConfig
        :return: None
        """
        assert isinstance(example_node_users_config.to_dict(), dict)
        assert isinstance(NodeUsersConfig.from_dict(example_node_users_config.to_dict()), NodeUsersConfig)
        assert NodeUsersConfig.from_dict(example_node_users_config.to_dict()).to_dict() == \
               example_node_users_config.to_dict()
        assert NodeUsersConfig.from_dict(example_node_users_config.to_dict()) == example_node_users_config

    def test_node_vulnerability_config(self, example_node_vulnerability_config: NodeVulnerabilityConfig) -> None:
        """
        Tests creation and dict conversion of the NodeVulnerabilitiesConfig DTO

        :param example_node_vulnerability_config: an example NodeVulnerabilityConfig
        :return: None
        """
        assert isinstance(example_node_vulnerability_config.to_dict(), dict)
        assert isinstance(NodeVulnerabilityConfig.from_dict(example_node_vulnerability_config.to_dict()),
                          NodeVulnerabilityConfig)
        assert NodeVulnerabilityConfig.from_dict(example_node_vulnerability_config.to_dict()).to_dict() == \
               example_node_vulnerability_config.to_dict()
        assert NodeVulnerabilityConfig.from_dict(example_node_vulnerability_config.to_dict()) == \
               example_node_vulnerability_config

    def test_ossec_ids_manager_config(self, example_ossec_ids_manager_config: OSSECIDSManagerConfig) -> None:
        """
        Tests creation and dict conversion of the OSSECIDSManagerConfig DTO

        :param example_ossec_ids_manager_config: an example OSSECIDSManagerConfig
        :return: None
        """
        assert isinstance(example_ossec_ids_manager_config.to_dict(), dict)
        assert isinstance(OSSECIDSManagerConfig.from_dict(example_ossec_ids_manager_config.to_dict()),
                          OSSECIDSManagerConfig)
        assert OSSECIDSManagerConfig.from_dict(example_ossec_ids_manager_config.to_dict()).to_dict() == \
               example_ossec_ids_manager_config.to_dict()
        assert OSSECIDSManagerConfig.from_dict(example_ossec_ids_manager_config.to_dict()) == \
               example_ossec_ids_manager_config

    def test_ovs_switch_config(self, example_ovs_switch_config: OvsSwitchConfig) -> None:
        """
        Tests creation and dict conversion of the OVSConfig DTO

        :param example_ovs_switch_config: an example OvsSwitchConfig
        :return: None
        """
        assert isinstance(example_ovs_switch_config.to_dict(), dict)
        assert isinstance(OvsSwitchConfig.from_dict(example_ovs_switch_config.to_dict()), OvsSwitchConfig)
        assert OvsSwitchConfig.from_dict(example_ovs_switch_config.to_dict()).to_dict() == \
               example_ovs_switch_config.to_dict()
        assert OvsSwitchConfig.from_dict(example_ovs_switch_config.to_dict()) == example_ovs_switch_config

    def test_ovs_config(self, example_ovs_config: OVSConfig) -> None:
        """
        Tests creation and dict conversion of the OVSConfig DTO

        :param example_ovs_config: an example OVSConfig
        :return: None
        """
        assert isinstance(example_ovs_config.to_dict(), dict)
        assert isinstance(OVSConfig.from_dict(example_ovs_config.to_dict()), OVSConfig)
        assert OVSConfig.from_dict(example_ovs_config.to_dict()).to_dict() == example_ovs_config.to_dict()
        assert OVSConfig.from_dict(example_ovs_config.to_dict()) == example_ovs_config

    def test_resources_config(self, example_resources_config: ResourcesConfig) -> None:
        """
        Tests creation and dict conversion of the ResourcesConfig DTO

        :param example_resources_config: an example ResourcesConfig
        :return: None
        """
        assert isinstance(example_resources_config.to_dict(), dict)
        assert isinstance(ResourcesConfig.from_dict(example_resources_config.to_dict()), ResourcesConfig)
        assert ResourcesConfig.from_dict(example_resources_config.to_dict()).to_dict() == \
               example_resources_config.to_dict()
        assert ResourcesConfig.from_dict(example_resources_config.to_dict()) == example_resources_config

    def test_sdn_controller_config(self, example_sdn_controller_config: SDNControllerConfig) -> None:
        """
        Tests creation and dict conversion of the ResourcesConfig DTO

        :param example_sdn_controller_config: an example SDNControllerConfig
        :return: None
        """
        assert isinstance(example_sdn_controller_config.to_dict(), dict)
        assert isinstance(SDNControllerConfig.from_dict(example_sdn_controller_config.to_dict()), SDNControllerConfig)
        assert SDNControllerConfig.from_dict(example_sdn_controller_config.to_dict()).to_dict() == \
               example_sdn_controller_config.to_dict()
        assert SDNControllerConfig.from_dict(example_sdn_controller_config.to_dict()) == example_sdn_controller_config

    def test_services_config(self, example_services_config: ServicesConfig) -> None:
        """
        Tests creation and dict conversion of the ServicesConfig DTO

        :param example_services_config: an example ServicesConfig
        :return: None
        """
        assert isinstance(example_services_config.to_dict(), dict)
        assert isinstance(ServicesConfig.from_dict(example_services_config.to_dict()), ServicesConfig)
        assert ServicesConfig.from_dict(
            example_services_config.to_dict()).to_dict() == example_services_config.to_dict()
        assert ServicesConfig.from_dict(example_services_config.to_dict()) == example_services_config

    def test_users_config(self, example_users_config: UsersConfig) -> None:
        """
        Tests creation and dict conversion of the UsersConfig DTO

        :param example_users_config: an example UsersConfig
        :return: None
        """
        assert isinstance(example_users_config.to_dict(), dict)
        assert isinstance(UsersConfig.from_dict(example_users_config.to_dict()), UsersConfig)
        assert UsersConfig.from_dict(example_users_config.to_dict()).to_dict() == example_users_config.to_dict()
        assert UsersConfig.from_dict(example_users_config.to_dict()) == example_users_config

    def test_arrival_config(self, example_constant_arrival_config: ConstantArrivalConfig) -> None:
        """
        Tests creation and dict conversion of the ArrivalConfig DTO

        :param example_constant_arrival_config: an example ConstantArrivalConfig
        :return: None
        """
        assert isinstance(example_constant_arrival_config.to_dict(), dict)
        assert isinstance(ConstantArrivalConfig.from_dict(example_constant_arrival_config.to_dict()),
                          ConstantArrivalConfig)
        assert ConstantArrivalConfig.from_dict(example_constant_arrival_config.to_dict()).to_dict() == \
               example_constant_arrival_config.to_dict()
        assert ConstantArrivalConfig.from_dict(example_constant_arrival_config.to_dict()) == \
               example_constant_arrival_config

    def test_client(self, example_client: Client) -> None:
        """
        Tests creation and dict conversion of the Client DTO

        :param example_client: an example Client
        :return: None
        """
        assert isinstance(example_client.to_dict(), dict)
        assert isinstance(Client.from_dict(example_client.to_dict()), Client)
        assert Client.from_dict(example_client.to_dict()).to_dict() == example_client.to_dict()
        assert Client.from_dict(example_client.to_dict()) == example_client

    def test_workflows_service(self, example_workflow_service: WorkflowService) -> None:
        """n
        Tests creation and dict conversion of the WorkflowService DTO

        :param example_workflow_service: an example WorkflowService
        :return: None
        """
        assert isinstance(example_workflow_service.to_dict(), dict)
        assert isinstance(WorkflowService.from_dict(example_workflow_service.to_dict()), WorkflowService)
        assert WorkflowService.from_dict(example_workflow_service.to_dict()).to_dict() == \
               example_workflow_service.to_dict()
        assert WorkflowService.from_dict(example_workflow_service.to_dict()) == example_workflow_service

    def test_workflow_markov_chain(self, example_workflow_markov_chain: WorkflowMarkovChain) -> None:
        """n
        Tests creation and dict conversion of the WorkflowMarkovChain DTO

        :param example_workflow_markov_chain: an example WorkflowMarkovChain
        :return: None
        """
        assert isinstance(example_workflow_markov_chain.to_dict(), dict)
        assert isinstance(WorkflowMarkovChain.from_dict(example_workflow_markov_chain.to_dict()), WorkflowMarkovChain)
        assert WorkflowMarkovChain.from_dict(example_workflow_markov_chain.to_dict()).to_dict() == \
               example_workflow_markov_chain.to_dict()
        assert WorkflowMarkovChain.from_dict(example_workflow_markov_chain.to_dict()) == example_workflow_markov_chain

    def test_workflows_config(self, example_workflows_config: WorkflowsConfig) -> None:
        """n
        Tests creation and dict conversion of the WorkflowsConfig DTO

        :param example_workflows_config: an example WorkflowsConfig
        :return: None
        """
        assert isinstance(example_workflows_config.to_dict(), dict)
        assert isinstance(WorkflowsConfig.from_dict(example_workflows_config.to_dict()), WorkflowsConfig)
        assert WorkflowsConfig.from_dict(example_workflows_config.to_dict()).to_dict() == \
               example_workflows_config.to_dict()
        assert WorkflowsConfig.from_dict(example_workflows_config.to_dict()) == example_workflows_config

    def test_client_population_config(self, example_client_population_config: ClientPopulationConfig) -> None:
        """
        Tests creation and dict conversion of the ClientPopulationConfig DTO

        :param example_client_population_config: an example ClientPopulationConfig
        :return: None
        """
        assert isinstance(example_client_population_config.to_dict(), dict)
        assert isinstance(ClientPopulationConfig.from_dict(example_client_population_config.to_dict()),
                          ClientPopulationConfig)
        assert ClientPopulationConfig.from_dict(example_client_population_config.to_dict()).to_dict() == \
               example_client_population_config.to_dict()
        assert ClientPopulationConfig.from_dict(example_client_population_config.to_dict()) == \
               example_client_population_config

    def test_traffic_config(self, example_traffic_config: TrafficConfig) -> None:
        """
        Tests creation and dict conversion of the TrafficConfig DTO

        :param example_traffic_config: an example TrafficConfig
        :return: None
        """
        assert isinstance(example_traffic_config.to_dict(), dict)
        assert isinstance(TrafficConfig.from_dict(example_traffic_config.to_dict()), TrafficConfig)
        assert TrafficConfig.from_dict(example_traffic_config.to_dict()).to_dict() == example_traffic_config.to_dict()
        assert TrafficConfig.from_dict(example_traffic_config.to_dict()) == example_traffic_config

    def test_vulnerabilities_config(self, example_vulnerabilities_config: VulnerabilitiesConfig) -> None:
        """
        Tests creation and dict conversion of the VulnerabilitiesConfig DTO

        :param example_vulnerabilities_config: an example VulnerabilitiesConfig
        :return: None
        """
        assert isinstance(example_vulnerabilities_config.to_dict(), dict)
        assert isinstance(VulnerabilitiesConfig.from_dict(example_vulnerabilities_config.to_dict()),
                          VulnerabilitiesConfig)
        assert VulnerabilitiesConfig.from_dict(example_vulnerabilities_config.to_dict()).to_dict() == \
               example_vulnerabilities_config.to_dict()
        assert VulnerabilitiesConfig.from_dict(example_vulnerabilities_config.to_dict()) == \
               example_vulnerabilities_config

    def test_emulation_connection_obs_state(
            self, example_emulation_connection_observation_state: EmulationConnectionObservationState) -> None:
        """
        Tests creation and dict conversion of the VulnerabilitiesConfig DTO

        :param example_emulation_connection_observation_state: an example EmulationConnectionObservationState
        :return: None
        """
        assert isinstance(example_emulation_connection_observation_state.to_dict(), dict)
        assert isinstance(
            EmulationConnectionObservationState.from_dict(example_emulation_connection_observation_state.to_dict()),
            EmulationConnectionObservationState)
        d1 = EmulationConnectionObservationState.from_dict(
            example_emulation_connection_observation_state.to_dict()).to_dict()
        d2 = example_emulation_connection_observation_state.to_dict()
        assert d1 == d2
        assert EmulationConnectionObservationState.from_dict(
            example_emulation_connection_observation_state.to_dict()) == example_emulation_connection_observation_state

    def test_connection_setup_dto(self, example_connection_setup_dto: ConnectionSetupDTO) -> None:
        """
        Tests creation and dict conversion of the ConnectionSetup DTO

        :param example_connection_setup_dto: an example ConnectionSetupDTO
        :return: None
        """
        assert isinstance(example_connection_setup_dto.to_dict(), dict)
        assert isinstance(ConnectionSetupDTO.from_dict(example_connection_setup_dto.to_dict()), ConnectionSetupDTO)
        assert ConnectionSetupDTO.from_dict(example_connection_setup_dto.to_dict()).to_dict() == \
               example_connection_setup_dto.to_dict()
        assert ConnectionSetupDTO.from_dict(example_connection_setup_dto.to_dict()) == example_connection_setup_dto

    def test_containers_config(self, example_containers_config: ContainersConfig) -> None:
        """
        Tests creation and dict conversion of the ContainersConfig DTO

        :param example_containers_config: an example ContainersConfig
        :return: None
        """
        assert isinstance(example_containers_config.to_dict(), dict)
        assert isinstance(ContainersConfig.from_dict(example_containers_config.to_dict()), ContainersConfig)
        assert ContainersConfig.from_dict(example_containers_config.to_dict()).to_dict() == \
               example_containers_config.to_dict()
        assert ContainersConfig.from_dict(example_containers_config.to_dict()) == example_containers_config

    def test_flags_config(self, example_flags_config: FlagsConfig) -> None:
        """
        Tests creation and dict conversion of the FlagsConfig DTO

        :param example_flags_config: an example FlagsConfig
        :return: None
        """
        assert isinstance(example_flags_config.to_dict(), dict)
        assert isinstance(FlagsConfig.from_dict(example_flags_config.to_dict()), FlagsConfig)
        assert FlagsConfig.from_dict(example_flags_config.to_dict()).to_dict() == example_flags_config.to_dict()
        assert FlagsConfig.from_dict(example_flags_config.to_dict()) == example_flags_config

    def test_topology_config(self, example_topology_config: TopologyConfig) -> None:
        """
        Tests creation and dict conversion of the TopologyConfig DTO

        :param example_topology_config: an example TopologyConfig
        :return: None
        """
        assert isinstance(example_topology_config.to_dict(), dict)
        assert isinstance(TopologyConfig.from_dict(example_topology_config.to_dict()), TopologyConfig)
        assert TopologyConfig.from_dict(example_topology_config.to_dict()).to_dict() == \
               example_topology_config.to_dict()
        assert TopologyConfig.from_dict(example_topology_config.to_dict()) == example_topology_config

    def test_host_manager_config(self, example_host_manager_config: HostManagerConfig) -> None:
        """
        Tests creation and dict conversion of the HostManagerConfig DTO

        :param example_host_manager_config: an example HostManagerConfig
        :return: None
        """
        assert isinstance(example_host_manager_config.to_dict(), dict)
        assert isinstance(HostManagerConfig.from_dict(example_host_manager_config.to_dict()), HostManagerConfig)
        assert HostManagerConfig.from_dict(example_host_manager_config.to_dict()).to_dict() == \
               example_host_manager_config.to_dict()
        assert HostManagerConfig.from_dict(example_host_manager_config.to_dict()) == example_host_manager_config

    def test_snort_ids_manager_config(self, example_snort_ids_manager_config: SnortIDSManagerConfig) -> None:
        """
        Tests creation and dict conversion of the SnortIDSManagerConfig DTO

        :param example_snort_ids_manager_config: an example SnortIDSManagerConfig
        :return: None
        """
        assert isinstance(example_snort_ids_manager_config.to_dict(), dict)
        assert isinstance(SnortIDSManagerConfig.from_dict(example_snort_ids_manager_config.to_dict()),
                          SnortIDSManagerConfig)
        assert SnortIDSManagerConfig.from_dict(example_snort_ids_manager_config.to_dict()).to_dict() == \
               example_snort_ids_manager_config.to_dict()
        assert SnortIDSManagerConfig.from_dict(example_snort_ids_manager_config.to_dict()) == \
               example_snort_ids_manager_config

    def test_emulation_env_config(self, example_emulation_env_config: EmulationEnvConfig) -> None:
        """
        Tests creation and dict conversion of the EmulationEnvConfig DTO

        :param example_emulation_env_config: an example EmulationEnvConfig
        :return: None
        """
        assert isinstance(example_emulation_env_config.to_dict(), dict)
        assert isinstance(EmulationEnvConfig.from_dict(example_emulation_env_config.to_dict()), EmulationEnvConfig)
        assert EmulationEnvConfig.from_dict(example_emulation_env_config.to_dict()).to_dict() == \
               example_emulation_env_config.to_dict()
        assert EmulationEnvConfig.from_dict(example_emulation_env_config.to_dict()) == example_emulation_env_config



from unittest.mock import patch, MagicMock
from csle_common.controllers.users_controller import UsersController


class TestUsersControllerSuite:
    """
    Test UsersController
    """

    @patch("csle_common.util.emulation_util.EmulationUtil.connect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.disconnect_admin")
    @patch("csle_common.util.emulation_util.EmulationUtil.execute_ssh_cmd")
    def test_create_users(self, mock_execute_ssh_cmd, mock_disconnect_admin, mock_connect_admin) -> None:
        """
        Test method that creates users in an emulation environment according to a specified users-configuration

        :param mock_execute_ssh_cmd: mock_execute_ssh_cmd
        :param mock_disconnect_admin: mock_disconnect_admin
        :param mock_connect_admin: mock_connect_admin
        :return: None
        """
        logger = MagicMock()
        user1 = MagicMock(username="user1", pw="password1", root=False)
        user2 = MagicMock(username="user2", pw="password2", root=True)
        users_conf = MagicMock(
            physical_host_ip="192.168.1.1", docker_gw_bridge_ip="192.168.1.2", users=[user1, user2])
        cred1 = MagicMock(username="cred1", pw="password1", root=False)
        cred2 = MagicMock(username="cred2", pw="password2", root=True)
        vuln_conf = MagicMock(
            physical_host_ip="192.168.1.1", docker_gw_bridge_ip="192.168.1.3", credentials=[cred1, cred2])
        emulation_env_config = MagicMock()
        emulation_env_config.users_config.users_configs = [users_conf]
        emulation_env_config.vuln_config.node_vulnerability_configs = [vuln_conf]
        emulation_env_config.connections = {"192.168.1.2": MagicMock(), "192.168.1.3": MagicMock()}

        physical_server_ip = "192.168.1.1"
        mock_execute_ssh_cmd.side_effect = [
            (b"user1\nuser2\n", "", 0),  # output of ls /home
            ("", "", 0),  # output of deluser user1
            ("", "", 0),  # output of rm -rf /home/user1
            ("", "", 0),  # output of deluser user2
            ("", "", 0),  # output of rm -rf /home/user2
            ("", "", 0),  # output of useradd for user1
            ("", "", 0),  # output of useradd for user2
            ("", "", 0),  # output of useradd for cred1
            ("", "", 0),  # output of useradd for cred2
        ]

        UsersController.create_users(emulation_env_config, physical_server_ip, logger)
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.2")
        mock_connect_admin.assert_any_call(emulation_env_config=emulation_env_config, ip="192.168.1.3")
        mock_execute_ssh_cmd.assert_called()
        mock_disconnect_admin.assert_called()



import logging
from typing import List
import time
import grpc
import csle_collector.client_manager.client_manager_pb2_grpc
import csle_collector.client_manager.client_manager_pb2
import csle_collector.client_manager.query_clients
import csle_collector.client_manager.client_manager_util
import csle_collector.traffic_manager.traffic_manager_pb2_grpc
import csle_collector.traffic_manager.traffic_manager_pb2
import csle_collector.traffic_manager.query_traffic_manager
import csle_collector.traffic_manager.traffic_manager_util
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
from csle_common.dao.emulation_config.emulation_env_config import EmulationEnvConfig
from csle_common.dao.emulation_config.client_managers_info import ClientManagersInfo
from csle_common.dao.emulation_config.node_traffic_config import NodeTrafficConfig
from csle_common.dao.emulation_config.traffic_managers_info import TrafficManagersInfo
from csle_common.dao.emulation_config.node_container_config import NodeContainerConfig
from csle_common.util.emulation_util import EmulationUtil


class TrafficController:
    """
    Class managing traffic generators in the emulation environments
    """

    @staticmethod
    def start_traffic_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                               logger: logging.Logger) -> None:
        """
        Utility method for checking if the traffic manager is running and starting it if it is not running
        on every node

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            if node_traffic_config.physical_host_ip != physical_server_ip:
                continue
            TrafficController.start_traffic_manager(emulation_env_config=emulation_env_config,
                                                    node_traffic_config=node_traffic_config,
                                                    logger=logger)

    @staticmethod
    def start_traffic_manager(emulation_env_config: EmulationEnvConfig,
                              node_traffic_config: NodeTrafficConfig,
                              logger: logging.Logger) -> None:
        """
        Utility method for starting traffic manager on a specific container

        :param emulation_env_config: the emulation env config
        :param node_traffic_config: traffic config of the container
        :param logger: the logger to use for logging
        :return: None
        """
        # Check if traffic_manager is already running
        status = None
        not_running = False
        try:
            status = TrafficController.get_traffic_manager_status_by_port_and_ip(
                ip=node_traffic_config.docker_gw_bridge_ip,
                port=node_traffic_config.traffic_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = f"running: {status.running}, script: {status.script}"
        if not_running:
            # Connect
            EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                        ip=node_traffic_config.docker_gw_bridge_ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.TRAFFIC_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=node_traffic_config.docker_gw_bridge_ip))

            # Start the _manager
            cmd = constants.COMMANDS.START_TRAFFIC_MANAGER.format(node_traffic_config.traffic_manager_port,
                                                                  node_traffic_config.traffic_manager_log_dir,
                                                                  node_traffic_config.traffic_manager_log_file,
                                                                  node_traffic_config.traffic_manager_max_workers)
            logger.info(f"Starting traffic manager on node "
                        f"{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip}), with cmd:{cmd}")
            o, e, _ = EmulationUtil.execute_ssh_cmd(
                cmd=cmd, conn=emulation_env_config.get_connection(ip=node_traffic_config.docker_gw_bridge_ip))
            time.sleep(10)
        else:
            logger.info(f"Traffic manager on node "
                        f"{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip}) was already running."
                        f"Status: {status_str}")

    @staticmethod
    def stop_traffic_managers(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                              logger: logging.Logger) -> None:
        """
        Utility method for stopping traffic managers on a given server

        :param emulation_env_config: the emulation env config
        :param physical_server_ip: the ip of the physical server to stop the traffic managers
        :param logger: the logger to use for logging
        :return: None
        """
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            if node_traffic_config.physical_host_ip == physical_server_ip:
                TrafficController.stop_traffic_manager(emulation_env_config=emulation_env_config,
                                                       node_traffic_config=node_traffic_config, logger=logger)

    @staticmethod
    def stop_traffic_manager(emulation_env_config: EmulationEnvConfig, node_traffic_config: NodeTrafficConfig,
                             logger: logging.Logger) -> None:
        """
        Utility method for stopping a specific traffic manager

        :param emulation_env_config: the emulation env config
        :param node_traffic_config: the node traffic configuration
        :param logger: the logger to use for logging
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=node_traffic_config.docker_gw_bridge_ip)
        logger.info(f"Stopping traffic manager on node {node_traffic_config.docker_gw_bridge_ip} "
                    f"({node_traffic_config.ip})")

        # Stop old background job if running
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.TRAFFIC_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd,
                                                conn=emulation_env_config.get_connection(
                                                    ip=node_traffic_config.docker_gw_bridge_ip))

    @staticmethod
    def start_client_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> bool:
        """
        Utility method starting the client manager

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: True if the client manager was started, False otherwise
        """
        # Check if client_manager is already running
        status = None
        not_running = False
        try:
            status = TrafficController.get_traffic_manager_status_by_port_and_ip(
                ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
                port=emulation_env_config.traffic_config.client_population_config.client_manager_port, timeout=5)
        except Exception:
            not_running = True
        status_str = ""
        if status is None:
            not_running = True
        else:
            status_str = (f"num_clients: {status.num_clients}, client_process_active: {status.client_process_active}, "
                          f"producer_active: {status.producer_active}, "
                          f"clients_time_step_len_seconds: {status.clients_time_step_len_seconds}, "
                          f"producer_time_step_len_seconds: {status.producer_time_step_len_seconds}")
        if not_running:
            logger.info(
                f"Starting client manager on container "
                f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                f"({emulation_env_config.traffic_config.client_population_config.ip})"
                f" since it was not running.")

            # Connect
            EmulationUtil.connect_admin(
                emulation_env_config=emulation_env_config,
                ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip)

            # Stop old background job if running
            cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
                   constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.CLIENT_MANAGER_FILE_NAME)
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(
                ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip))

            # Start the client_manager
            cmd = constants.COMMANDS.START_CLIENT_MANAGER.format(
                emulation_env_config.traffic_config.client_population_config.client_manager_port,
                emulation_env_config.traffic_config.client_population_config.client_manager_log_dir,
                emulation_env_config.traffic_config.client_population_config.client_manager_log_file,
                emulation_env_config.traffic_config.client_population_config.client_manager_max_workers
            )
            logger.info(
                f"Starting client manager on container "
                f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                f"with cmd: {cmd}")
            o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(
                ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip))
            time.sleep(10)
            return True
        else:
            logger.info(
                f"Client manager was already running on container "
                f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                f"({emulation_env_config.traffic_config.client_population_config.ip}). Status: {status_str}")
            return False

    @staticmethod
    def stop_client_manager(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Utility method starting the client manager

        :param emulation_env_config: the emulation env config
        :return: None
        """
        # Connect
        EmulationUtil.connect_admin(emulation_env_config=emulation_env_config,
                                    ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip)

        logger.info(f"Stopping client manager on node "
                    f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                    f"({emulation_env_config.traffic_config.client_population_config.ip})")

        # Stop old background job if running
        cmd = (constants.COMMANDS.SUDO + constants.COMMANDS.SPACE_DELIM + constants.COMMANDS.PKILL +
               constants.COMMANDS.SPACE_DELIM + constants.TRAFFIC_COMMANDS.CLIENT_MANAGER_FILE_NAME)
        o, e, _ = EmulationUtil.execute_ssh_cmd(cmd=cmd, conn=emulation_env_config.get_connection(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip))

    @staticmethod
    def stop_client_population(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) -> None:
        """
        Function for stopping the client arrival process of an emulation

        :param emulation_env_config: the emulation env config
        :param logger: the logger to use for logging
        :return: None
        """
        logger.info(
            f"Stopping client population on container: "
            f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({emulation_env_config.traffic_config.client_population_config.ip})")

        TrafficController.start_client_manager(emulation_env_config=emulation_env_config, logger=logger)

        client_dto = TrafficController.get_clients_dto_by_ip_and_port(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
            port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
            logger=logger)

        if client_dto.client_process_active:
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip}:'
                    f'{emulation_env_config.traffic_config.client_population_config.client_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.client_manager.client_manager_pb2_grpc.ClientManagerStub(channel)
                csle_collector.client_manager.query_clients.stop_clients(stub)

    @staticmethod
    def start_client_producer(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                              logger: logging.Logger) -> None:
        """
        Starts the Kafka producer for client metrics

        :param emulation_env_config: the emulation environment configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.traffic_config.client_population_config.physical_host_ip != physical_server_ip:
            return
        logger.info(
            f"Starting client producer on container:"
            f" {emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({emulation_env_config.traffic_config.client_population_config.ip})")
        client_manager_started = (
            TrafficController.start_client_manager(emulation_env_config=emulation_env_config, logger=logger))

        if client_manager_started:
            logger.info(
                f"Client manager was not running on container:"
                f" {emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                f"({emulation_env_config.traffic_config.client_population_config.ip}), started it")
            time.sleep(10)
            # If client manager was started we need to first start the client population
            TrafficController.start_client_population(emulation_env_config=emulation_env_config,
                                                      physical_server_ip=physical_server_ip, logger=logger)
        time.sleep(5)
        client_dto = TrafficController.get_clients_dto_by_ip_and_port(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
            port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
            logger=logger)
        if not client_dto.producer_active:
            time.sleep(5)
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip}:'
                    f'{emulation_env_config.traffic_config.client_population_config.client_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.client_manager.client_manager_pb2_grpc.ClientManagerStub(channel)

                # Start the producer thread
                csle_collector.client_manager.query_clients.start_producer(
                    stub=stub, ip=emulation_env_config.kafka_config.container.get_ips()[0],
                    port=emulation_env_config.kafka_config.kafka_port,
                    time_step_len_seconds=emulation_env_config.kafka_config.time_step_len_seconds)

    @staticmethod
    def stop_client_producer(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                             logger: logging.Logger) -> None:
        """
        Stops the Kafka producer for client metrics

        :param emulation_env_config: the emulation environment configuration
        :param physical_server_ip: ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.traffic_config.client_population_config.physical_host_ip != physical_server_ip:
            return

        logger.info(
            f"Stopping client producer on container:"
            f" {emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
            f"({emulation_env_config.traffic_config.client_population_config.ip})")

        TrafficController.start_client_manager(emulation_env_config=emulation_env_config, logger=logger)

        client_dto = TrafficController.get_clients_dto_by_ip_and_port(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
            port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
            logger=logger)

        if client_dto.producer_active:
            # Open a gRPC session
            with grpc.insecure_channel(
                    f'{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip}:'
                    f'{emulation_env_config.traffic_config.client_population_config.client_manager_port}',
                    options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
                stub = csle_collector.client_manager.client_manager_pb2_grpc.ClientManagerStub(channel)

                # Stop the producer thread
                csle_collector.client_manager.query_clients.stop_producer(stub=stub)

    @staticmethod
    def start_client_population(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                logger: logging.Logger) -> None:
        """
        Starts the arrival process of clients

        :param emulation_env_config: the emulation environment configuration
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        if emulation_env_config.traffic_config.client_population_config.physical_host_ip != physical_server_ip:
            return
        logger.info(f"Starting client population on container: "
                    f"{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip} "
                    f"({emulation_env_config.traffic_config.client_population_config.ip})")

        TrafficController.start_client_manager(emulation_env_config=emulation_env_config, logger=logger)

        client_dto = TrafficController.get_clients_dto_by_ip_and_port(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
            port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
            logger=logger)

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip}:'
                f'{emulation_env_config.traffic_config.client_population_config.client_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.client_manager.client_manager_pb2_grpc.ClientManagerStub(channel)

            # Stop the client population if it is already running
            if client_dto.client_process_active:
                csle_collector.client_manager.query_clients.stop_clients(stub)
                time.sleep(5)

            # Start the client population
            time_step_len = emulation_env_config.traffic_config.client_population_config.client_time_step_len_seconds
            num_workflows = len(
                emulation_env_config.traffic_config.client_population_config.workflows_config.workflow_markov_chains)
            num_services = len(
                emulation_env_config.traffic_config.client_population_config.workflows_config.workflow_services)
            logger.info(
                f"Starting the client population, "
                f"time_step_len_seconds: {time_step_len},\n"
                f"num client profiles: {len(emulation_env_config.traffic_config.client_population_config.clients)}, \n"
                f"num workflows: {num_workflows}, num_services: {num_services}")
            csle_collector.client_manager.query_clients.start_clients(
                stub=stub, time_step_len_seconds=time_step_len,
                workflows_config=emulation_env_config.traffic_config.client_population_config.workflows_config,
                clients=emulation_env_config.traffic_config.client_population_config.clients)

    @staticmethod
    def get_num_active_clients(emulation_env_config: EmulationEnvConfig, logger: logging.Logger) \
            -> csle_collector.client_manager.client_manager_pb2.ClientsDTO:
        """
        Gets the number of active clients

        :param emulation_env_config: the emulation configuration
        :param logger: the logger to use for logging
        :return: A ClientDTO which contains the number of active clients
        """

        TrafficController.start_client_manager(emulation_env_config=emulation_env_config, logger=logger)

        client_dto = TrafficController.get_clients_dto_by_ip_and_port(
            ip=emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip,
            port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
            logger=logger)
        return client_dto

    @staticmethod
    def get_clients_dto_by_ip_and_port(ip: str, port: int, logger: logging.Logger,
                                       timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) -> \
            csle_collector.client_manager.client_manager_pb2.ClientsDTO:
        """
        A method that sends a request to the ClientManager on a specific container
        to get its status

        :param ip: the ip of the container
        :param port: the port of the client manager running on the container
        :param logger: the logger to use for logging
        :param timeout: the timeout of the GRPC query
        :return: the status of the clientmanager
        """
        logger.info(f"Get client manager status from container with ip: {ip} and client manager port: {port}")
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.client_manager.client_manager_pb2_grpc.ClientManagerStub(channel)
            status = csle_collector.client_manager.query_clients.get_clients(stub=stub, timeout=timeout)
            return status

    @staticmethod
    def stop_internal_traffic_generators(emulation_env_config: EmulationEnvConfig, logger: logging.Logger,
                                         physical_server_ip: str) -> None:
        """
        Utility function for stopping internal traffic generators

        :param emulation_env_config: the configuration of the emulation env
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            if node_traffic_config.physical_host_ip == physical_server_ip:
                TrafficController.stop_internal_traffic_generator(emulation_env_config=emulation_env_config,
                                                                  node_traffic_config=node_traffic_config,
                                                                  logger=logger)

    @staticmethod
    def stop_internal_traffic_generator(emulation_env_config: EmulationEnvConfig,
                                        node_traffic_config: NodeTrafficConfig, logger: logging.Logger) -> None:
        """
        Utility function for stopping a specific internal traffic generator

        :param emulation_env_config: the configuration of the emulation env
        :param logger: the logger to use for logging
        :param node_traffic_config: the node traffic config
        :return: None
        """
        TrafficController.start_traffic_manager(emulation_env_config=emulation_env_config,
                                                node_traffic_config=node_traffic_config, logger=logger)
        logger.info(f"Stopping traffic generator script, "
                    f"node ip:{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip})")

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{node_traffic_config.docker_gw_bridge_ip}:{node_traffic_config.traffic_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.traffic_manager.traffic_manager_pb2_grpc.TrafficManagerStub(channel)
            csle_collector.traffic_manager.query_traffic_manager.stop_traffic(stub)

    @staticmethod
    def start_internal_traffic_generators(emulation_env_config: EmulationEnvConfig, physical_server_ip: str,
                                          logger: logging.Logger) -> None:
        """
        Utility function for starting internal traffic generators

        :param emulation_env_config: the configuration of the emulation env
        :param physical_server_ip: the ip of the physical server
        :param logger: the logger to use for logging
        :return: None
        """
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            if node_traffic_config.physical_host_ip != physical_server_ip:
                continue
            container = None
            for c in emulation_env_config.containers_config.containers:
                if node_traffic_config.ip in c.get_ips():
                    container = c
                    break
            if container is None:
                continue
            else:
                TrafficController.start_internal_traffic_generator(
                    emulation_env_config=emulation_env_config, node_traffic_config=node_traffic_config,
                    container=container, logger=logger)

    @staticmethod
    def start_internal_traffic_generator(
            emulation_env_config: EmulationEnvConfig, node_traffic_config: NodeTrafficConfig,
            container: NodeContainerConfig, logger: logging.Logger) -> None:
        """
        Utility function for starting internal traffic generators

        :param emulation_env_config: the configuration of the emulation env
        :param node_traffic_config: the node traffic configuration
        :param container: the container
        :param logger: the logger to use for logging
        :return: None
        """
        TrafficController.start_traffic_manager(emulation_env_config=emulation_env_config,
                                                node_traffic_config=node_traffic_config, logger=logger)
        logger.info(f"Starting traffic generator script, "
                    f"node ip:{node_traffic_config.docker_gw_bridge_ip} ({node_traffic_config.ip})")

        commands = []
        subnet_masks = []
        reachable_containers = []
        for ip_net1 in container.ips_and_networks:
            ip, net1 = ip_net1
            subnet_masks.append(net1.subnet_mask)
        for c in emulation_env_config.containers_config.containers:
            for ip_net1 in c.ips_and_networks:
                ip, net1 = ip_net1
                if net1.subnet_mask in subnet_masks and ip not in container.get_ips():
                    reachable_containers.append((ip, c.get_ips()))

        for node2 in emulation_env_config.traffic_config.node_traffic_configs:
            for rc in reachable_containers:
                ip, ips = rc
                if node2.ip in ips:
                    for cmd in node2.commands:
                        commands.append(cmd.format(ip))

        sleep_time = emulation_env_config.traffic_config.client_population_config.client_time_step_len_seconds

        # Open a gRPC session
        with grpc.insecure_channel(
                f'{node_traffic_config.docker_gw_bridge_ip}:{node_traffic_config.traffic_manager_port}',
                options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.traffic_manager.traffic_manager_pb2_grpc.TrafficManagerStub(channel)
            csle_collector.traffic_manager.query_traffic_manager.start_traffic(stub, commands=commands,
                                                                               sleep_time=sleep_time)

    @staticmethod
    def get_client_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the IPS of the Client managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        return [emulation_env_config.traffic_config.client_population_config.docker_gw_bridge_ip]

    @staticmethod
    def get_client_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Client managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        return [emulation_env_config.traffic_config.client_population_config.client_manager_port]

    @staticmethod
    def get_client_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                 logger: logging.Logger) -> ClientManagersInfo:
        """
        Extracts the information of the Client managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param logger: the logger to use for logging
        :return: a DTO with the status of the Client managers
        """
        client_managers_ips = TrafficController.get_client_managers_ips(emulation_env_config=emulation_env_config)
        client_managers_ports = TrafficController.get_client_managers_ports(emulation_env_config=emulation_env_config)
        client_managers_statuses = []
        client_managers_running = []
        for ip in client_managers_ips:
            if ip not in active_ips:
                continue
            running = False
            status = None
            try:
                status = TrafficController.get_clients_dto_by_ip_and_port(
                    ip=ip, port=emulation_env_config.traffic_config.client_population_config.client_manager_port,
                    logger=logger)
                running = True
            except Exception as e:
                logger.info(
                    f"Could not fetch client manager status on IP:{ip}, error: {str(e)}, {repr(e)}")
            if status is not None:
                client_managers_statuses.append(status)
            else:
                client_managers_statuses.append(
                    csle_collector.client_manager.client_manager_util.ClientManagerUtil.clients_dto_empty())
            client_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        client_manager_info_dto = ClientManagersInfo(
            client_managers_running=client_managers_running, ips=client_managers_ips, execution_id=execution_id,
            emulation_name=emulation_name, client_managers_statuses=client_managers_statuses,
            ports=client_managers_ports)
        return client_manager_info_dto

    @staticmethod
    def get_traffic_manager_status_by_port_and_ip(
            ip: str, port: int, timeout: int = collector_constants.GRPC.TIMEOUT_SECONDS) -> \
            csle_collector.traffic_manager.traffic_manager_pb2.TrafficDTO:
        """
        A method that sends a request to the TrafficManager on a specific container
        to get its status

        :param ip: the ip where the traffic manager is running
        :param port: the port the traffic manager is listening to
        :param timeout: the timeout of the GRPC query
        :return: the status of the traffic manager
        """
        # Open a gRPC session
        with grpc.insecure_channel(f'{ip}:{port}', options=constants.GRPC_SERVERS.GRPC_OPTIONS) as channel:
            stub = csle_collector.traffic_manager.traffic_manager_pb2_grpc.TrafficManagerStub(channel)
            status = csle_collector.traffic_manager.query_traffic_manager.get_traffic_status(stub=stub, timeout=timeout)
            return status

    @staticmethod
    def get_traffic_managers_ips(emulation_env_config: EmulationEnvConfig) -> List[str]:
        """
        A method that extracts the ips of the traffic managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of IP addresses
        """
        ips = []
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            ips.append(node_traffic_config.docker_gw_bridge_ip)
        return ips

    @staticmethod
    def get_traffic_managers_ports(emulation_env_config: EmulationEnvConfig) -> List[int]:
        """
        A method that extracts the ports of the Traffic managers in a given emulation

        :param emulation_env_config: the emulation env config
        :return: the list of ports
        """
        ports = []
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            ports.append(node_traffic_config.traffic_manager_port)
        return ports

    @staticmethod
    def get_traffic_managers_info(emulation_env_config: EmulationEnvConfig, active_ips: List[str],
                                  physical_host_ip: str, logger: logging.Logger) \
            -> TrafficManagersInfo:
        """
        Extracts the information of the traffic managers for a given emulation

        :param emulation_env_config: the configuration of the emulation
        :param active_ips: list of active IPs
        :param physical_host_ip: the ip of the physical host
        :param logger: the logger to use for logging
        :return: a DTO with the status of the traffic managers
        """
        traffic_managers_ips = TrafficController.get_traffic_managers_ips(emulation_env_config=emulation_env_config)
        traffic_managers_ports = TrafficController.get_traffic_managers_ports(emulation_env_config=emulation_env_config)
        traffic_managers_statuses = []
        traffic_managers_running = []
        for node_traffic_config in emulation_env_config.traffic_config.node_traffic_configs:
            logger.info(f"Getting traffic information from node: {node_traffic_config.ip}")
            if node_traffic_config.docker_gw_bridge_ip not in active_ips or not EmulationUtil.physical_ip_match(
                    emulation_env_config=emulation_env_config, ip=node_traffic_config.docker_gw_bridge_ip,
                    physical_host_ip=physical_host_ip):
                continue
            running = False
            status = None
            if node_traffic_config.docker_gw_bridge_ip in traffic_managers_ips:
                try:
                    status = TrafficController.get_traffic_manager_status_by_port_and_ip(
                        port=node_traffic_config.traffic_manager_port, ip=node_traffic_config.docker_gw_bridge_ip)
                    running = True
                except Exception as e:
                    logger.info(f"Could not fetch traffic manager status on IP"
                                f":{node_traffic_config.docker_gw_bridge_ip}:"
                                f"{node_traffic_config.traffic_manager_port}, error: {str(e)}, {repr(e)}")
                if status is not None:
                    traffic_managers_statuses.append(status)
                else:
                    traffic_managers_statuses.append(
                        csle_collector.traffic_manager.traffic_manager_util.TrafficManagerUtil.traffic_dto_empty())
                traffic_managers_running.append(running)
        execution_id = emulation_env_config.execution_id
        emulation_name = emulation_env_config.name
        traffic_manager_info_dto = TrafficManagersInfo(
            traffic_managers_running=traffic_managers_running, ips=traffic_managers_ips, execution_id=execution_id,
            emulation_name=emulation_name, traffic_managers_statuses=traffic_managers_statuses,
            ports=traffic_managers_ports)
        return traffic_manager_info_dto



from typing import Dict, Any, List
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.emulation_config.emulation_trace import EmulationTrace
from csle_base.json_serializable import JSONSerializable


class DataCollectionJobConfig(JSONSerializable):
    """
    DTO representing the configuration of a data collection job
    """

    def __init__(self, emulation_env_name: str, num_collected_steps: int,
                 progress_percentage: float, attacker_sequence: List[EmulationAttackerAction], pid: int,
                 repeat_times: int, emulation_statistic_id: int, num_sequences_completed: int,
                 traces: List[EmulationTrace], save_emulation_traces_every: int, num_cached_traces: int,
                 defender_sequence: List[EmulationDefenderAction], log_file_path: str, physical_host_ip: str,
                 descr: str = ""):
        """
        Initializes the DTO

        :param emulation_env_name: the emulation environment name
        :param num_collected_steps: number of collected steps in the emulation
        :param num_sequences_completed: number of sequences completed
        :param progress_percentage: the progress of the data collection job in %
        :param attacker_sequence: the sequence of actions to emulate the attacker
        :param defender_sequence: the sequence of actions to emulate the defender
        :param repeat_times: the number of times to repeat the sequences
        :param traces: list of collected emulation traces
        :param descr: description of the job
        :param emulation_statistic_id: the id of the emulation statistic
        :param save_emulation_traces_every: the frequency to save emulation traces to the metastore
        :param num_cached_traces: the number of emulation traces to keep with the job metadata
        :param physical_host_ip: the IP of the physical host where the job is running
        """
        self.emulation_env_name = emulation_env_name
        self.progress_percentage = round(progress_percentage, 3)
        self.pid = pid
        self.num_collected_steps = num_collected_steps
        self.progress_percentage = self.progress_percentage
        self.attacker_sequence = attacker_sequence
        self.defender_sequence = defender_sequence
        self.id = -1
        self.running = False
        self.descr = descr
        self.repeat_times = repeat_times
        self.emulation_statistic_id = emulation_statistic_id
        self.num_sequences_completed = num_sequences_completed
        self.traces = traces
        self.save_emulation_traces_every = save_emulation_traces_every
        self.num_cached_traces = num_cached_traces
        self.log_file_path = log_file_path
        self.physical_host_ip = physical_host_ip

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["emulation_env_name"] = self.emulation_env_name
        d["progress_percentage"] = round(self.progress_percentage, 2)
        d["pid"] = self.pid
        d["num_collected_steps"] = self.num_collected_steps
        d["progress_percentage"] = self.progress_percentage
        d["attacker_sequence"] = list(map(lambda x: x.to_dict(), self.attacker_sequence))
        d["defender_sequence"] = list(map(lambda x: x.to_dict(), self.defender_sequence))
        d["id"] = self.id
        d["running"] = self.running
        d["descr"] = self.descr
        d["repeat_times"] = self.repeat_times
        d["emulation_statistic_id"] = self.emulation_statistic_id
        d["traces"] = list(map(lambda x: x.to_dict(), self.traces))
        d["num_sequences_completed"] = self.num_sequences_completed
        d["save_emulation_traces_every"] = self.save_emulation_traces_every
        d["num_cached_traces"] = self.num_cached_traces
        d["log_file_path"] = self.log_file_path
        d["physical_host_ip"] = self.physical_host_ip
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "DataCollectionJobConfig":
        """
        Converts a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = DataCollectionJobConfig(
            emulation_env_name=d["emulation_env_name"], pid=d["pid"], num_collected_steps=d["num_collected_steps"],
            progress_percentage=d["progress_percentage"],
            attacker_sequence=list(map(lambda x: EmulationAttackerAction.from_dict(x), d["attacker_sequence"])),
            defender_sequence=list(map(lambda x: EmulationDefenderAction.from_dict(x), d["defender_sequence"])),
            descr=d["descr"], repeat_times=d["repeat_times"], emulation_statistic_id=d["emulation_statistic_id"],
            traces=list(map(lambda x: EmulationTrace.from_dict(x), d["traces"])),
            num_sequences_completed=d["num_sequences_completed"],
            save_emulation_traces_every=d["save_emulation_traces_every"], num_cached_traces=d["num_cached_traces"],
            log_file_path=d["log_file_path"], physical_host_ip=d["physical_host_ip"])
        obj.id = d["id"]
        obj.running = d["running"]
        return obj

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"emulation_env_name: {self.emulation_env_name}, pid: {self.pid}, " \
               f"progress_percentage: {self.progress_percentage}, " \
               f"attacker_sequence={list(map(lambda x: str(x), self.attacker_sequence))}," \
               f"defender_sequence={list(map(lambda x: str(x), self.defender_sequence))}, id: {self.id}," \
               f"running:{self.running}, descr: {self.descr}, repeat_times: {self.repeat_times}," \
               f"emulation_statistic_id: {self.emulation_statistic_id}, " \
               f"num_sequences_completed: {self.num_sequences_completed}, " \
               f"traces: {list(map(lambda x: str(x), self.traces))}, " \
               f"save_emulation_traces_every: {self.save_emulation_traces_every}, " \
               f"num_cached_traces: {self.num_cached_traces}, log_file_path: {self.log_file_path}, " \
               f"physical_host_ip: {self.physical_host_ip}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "DataCollectionJobConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return DataCollectionJobConfig.from_dict(json.loads(json_str))



from typing import List, Dict, Any
import csle_collector.ryu_manager.ryu_manager_pb2_grpc
import csle_collector.ryu_manager.ryu_manager_pb2
import csle_collector.ryu_manager.ryu_manager_util as ryu_manager_util
from csle_base.json_serializable import JSONSerializable


class RyuManagersInfo(JSONSerializable):
    """
    DTO containing the status of the Ryu managers for a given emulation execution
    """

    def __init__(self, ips: List[str], ports: List[int],
                 emulation_name: str, execution_id: int,
                 ryu_managers_statuses: List[csle_collector.ryu_manager.ryu_manager_pb2.RyuDTO],
                 ryu_managers_running: List[bool], local_controller_web_port: int = -1,
                 physical_server_ip: str = ""):
        """
        Initializes the DTO

        :param ryu_managers_running: boolean list that indicate whether the Ryu managers are running
        :param ips: the list of IPs of the running Ryu managers
        :param ports: the list of ports of the running Ryu managers
        :param emulation_name: the name of the corresponding emulation
        :param execution_id: the ID of the corresponding emulation execution
        :param ryu_managers_statuses: a list of statuses of the Ryu managers
        :param local_controller_web_port: the local port that is forwarded to the SDN controller's web api
        :param physical_server_ip: the ip of the physical server where Ryu is running
        """
        self.ryu_managers_running = ryu_managers_running
        self.ips = ips
        self.ports = ports
        self.emulation_name = emulation_name
        self.execution_id = execution_id
        self.ryu_managers_statuses = ryu_managers_statuses
        self.local_controller_web_port = local_controller_web_port
        self.physical_server_ip = physical_server_ip

    def __str__(self):
        """
        :return: a string representation of the DTO
        """
        return f"ryu_managers_running: {self.ryu_managers_running}, ips: {list(map(lambda x: str(x), self.ips))}, " \
               f"emulation_name: {self.emulation_name}, " \
               f"execution_id: {self.execution_id}, " \
               f"ryu_managers_statuses: {list(map(lambda x: str(x), self.ryu_managers_statuses))}, " \
               f"ports: {list(map(lambda x: str(x), self.ports))}," \
               f"local_controller_web_port: {self.local_controller_web_port}, " \
               f"physical_server_ip: {self.physical_server_ip}"

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["ryu_managers_running"] = self.ryu_managers_running
        d["ips"] = self.ips
        d["ports"] = self.ports
        d["emulation_name"] = self.emulation_name
        d["execution_id"] = self.execution_id
        d["local_controller_web_port"] = self.local_controller_web_port
        d["physical_server_ip"] = self.physical_server_ip
        d["ryu_managers_statuses"] = list(map(
            lambda x: ryu_manager_util.RyuManagerUtil.ryu_dto_to_dict(x), self.ryu_managers_statuses))
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "RyuManagersInfo":
        """
        Convert a dict representation to a DTO representation

        :return: a dto representation of the object
        """
        dto = RyuManagersInfo(ryu_managers_running=d["ryu_managers_running"], ips=d["ips"], ports=d["ports"],
                              emulation_name=d["emulation_name"], execution_id=d["execution_id"],
                              ryu_managers_statuses=list(map(
                                  lambda x: ryu_manager_util.RyuManagerUtil.ryu_dto_from_dict(x),
                                  d["ryu_managers_statuses"])),
                              local_controller_web_port=d["local_controller_web_port"],
                              physical_server_ip=d["physical_server_ip"])
        return dto

    @staticmethod
    def from_json_file(json_file_path: str) -> "RyuManagersInfo":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return RyuManagersInfo.from_dict(json.loads(json_str))



from typing import Optional, List, Dict, Any, Tuple, Union
import json
import os
from csle_common.logging.log import Logger
import csle_common.constants.constants as constants
import csle_collector.constants.constants as collector_constants
from csle_common.dao.emulation_action.attacker.emulation_attacker_action import EmulationAttackerAction
from csle_common.dao.emulation_action.attacker.emulation_attacker_action_id import EmulationAttackerActionId
from csle_common.dao.emulation_action.defender.emulation_defender_action import EmulationDefenderAction
from csle_common.dao.encoding.np_encoder import NpEncoder
from csle_common.dao.emulation_observation.attacker.emulation_attacker_observation_state \
    import EmulationAttackerObservationState
from csle_common.dao.emulation_observation.defender.emulation_defender_observation_state \
    import EmulationDefenderObservationState
from csle_base.json_serializable import JSONSerializable


class EmulationTrace(JSONSerializable):
    """
    DTO class representing a trace in the emulation system
    """

    def __init__(self, initial_attacker_observation_state: EmulationAttackerObservationState,
                 initial_defender_observation_state: EmulationDefenderObservationState, emulation_name: str):
        """
        Initializes the DTO

        :param initial_attacker_observation_state: the initial state of the attacker
        :param initial_defender_observation_state: the intial state of the defender
        :param emulation_name: the name of the emulation
        """
        self.initial_attacker_observation_state = initial_attacker_observation_state
        self.initial_defender_observation_state = initial_defender_observation_state
        self.attacker_observation_states: List[EmulationAttackerObservationState] = []
        self.defender_observation_states: List[EmulationDefenderObservationState] = []
        self.attacker_actions: List[EmulationAttackerAction] = []
        self.defender_actions: List[EmulationDefenderAction] = []
        self.emulation_name = emulation_name
        self.id = -1

    def __str__(self):
        """
        :return: a string representation of the object
        """
        return f"initial_attacker_observation_state:{self.initial_attacker_observation_state}" \
               f"initial_defender_observation_state:{self.initial_defender_observation_state}" \
               f"attacker_observation_states:{self.attacker_observation_states}\n" \
               f"defender_observation_states:{self.defender_observation_states}\n" \
               f"attacker_actions:{self.attacker_actions}\n" \
               f"defender_actions:{self.defender_actions}\n" \
               f"emulation_name: {self.emulation_name}," \
               f"id:{self.id}"

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "EmulationTrace":
        """
        Converts a dict representation into an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = EmulationTrace(
            initial_attacker_observation_state=EmulationAttackerObservationState.from_dict(
                d["initial_attacker_observation_state"]),
            initial_defender_observation_state=EmulationDefenderObservationState.from_dict(
                d["initial_defender_observation_state"]),
            emulation_name=d["emulation_name"]
        )
        obj.attacker_observation_states = list(map(lambda x: EmulationAttackerObservationState.from_dict(x),
                                                   d["attacker_observation_states"]))
        obj.defender_observation_states = list(map(lambda x: EmulationDefenderObservationState.from_dict(x),
                                                   d["defender_observation_states"]))
        obj.attacker_actions = list(map(lambda x: EmulationAttackerAction.from_dict(x), d["attacker_actions"]))
        obj.defender_actions = list(map(lambda x: EmulationDefenderAction.from_dict(x),
                                        d["defender_actions"]))
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["initial_attacker_observation_state"] = self.initial_attacker_observation_state.to_dict()
        d["initial_defender_observation_state"] = self.initial_defender_observation_state.to_dict()
        d["attacker_observation_states"] = list(map(lambda x: x.to_dict(), self.attacker_observation_states))
        d["defender_observation_states"] = list(map(lambda x: x.to_dict(), self.defender_observation_states))
        d["attacker_actions"] = list(map(lambda x: x.to_dict(), self.attacker_actions))
        d["defender_actions"] = list(map(lambda x: x.to_dict(), self.defender_actions))
        d["emulation_name"] = self.emulation_name
        d["id"] = self.id
        return d

    @staticmethod
    def save_traces_to_disk(traces_save_dir, traces: List["EmulationTrace"], traces_file: Optional[str] = None) -> None:
        """
        Utility function for saving a list of traces to a json file

        :param traces_save_dir: the directory where to save the traces
        :param traces: the traces to save
        :param traces_file: the filename of the traces file
        :return: None
        """
        traces_1 = list(map(lambda x: x.to_dict(), traces))
        if traces_file is None:
            traces_file = constants.SYSTEM_IDENTIFICATION.EMULATION_TRACES_FILE
        if not os.path.exists(traces_save_dir):
            os.makedirs(traces_save_dir)
        with open(traces_save_dir + "/" + traces_file, 'w') as fp:
            json.dump({"traces": traces_1}, fp, cls=NpEncoder)

    @staticmethod
    def load_traces_from_disk(traces_file: str) -> List["EmulationTrace"]:
        """
        Utility function for loading and parsing a list of traces from a json file

        :param traces_file: (optional) a custom name of the traces file
        :return: a list of the loaded traces
        """
        if os.path.exists(traces_file):
            with open(traces_file, 'r') as fp:
                d = json.load(fp)
                if constants.METADATA_STORE.TRACES_PROPERTY in d:
                    traces = d[constants.METADATA_STORE.TRACES_PROPERTY]
                else:
                    traces = d["emulations"]
                traces_1 = list(map(lambda x: EmulationTrace.from_dict(x), traces))
                return traces_1
        else:
            Logger.__call__().get_logger().info("Warning: Could not "
                                                f"read traces file, path does not exist:{traces_file}")
            return []

    @staticmethod
    def from_json_str(json_str: str) -> "EmulationTrace":
        """
        Converts json string into a DTO

        :param json_str: the json string representation
        :return: the DTO instance
        """
        import json
        dto: EmulationTrace = EmulationTrace.from_dict(json.loads(json_str))
        return dto

    @staticmethod
    def from_json_file(json_file_path: str) -> "EmulationTrace":
        """
        Reads a json file and converts it into a dto

        :param json_file_path: the json file path to save  the DTO to
        :return: None
        """
        import io
        with io.open(json_file_path, 'r', encoding='utf-8') as f:
            json_str = f.read()
            dto = EmulationTrace.from_json_str(json_str=json_str)
            return dto

    def num_attributes_per_time_step(self) -> int:
        """
        :return: approximately the number of attributes recorded per time-step of the trace
        """
        num_attributes: int = 2
        num_attributes = (num_attributes + (1 + len(self.attacker_observation_states))
                          * self.initial_attacker_observation_state.num_attributes())
        num_attributes = (num_attributes + (1 + len(self.defender_observation_states))
                          * self.initial_defender_observation_state.num_attributes())
        if len(self.defender_actions) > 0:
            num_attributes = num_attributes + len(self.defender_actions) * self.defender_actions[0].num_attributes()
        if len(self.attacker_actions) > 0:
            num_attributes = num_attributes + len(self.attacker_actions) * self.attacker_actions[0].num_attributes()
        return num_attributes

    @staticmethod
    def schema():
        """
        :return: the schema of the DTO
        """
        dto = EmulationTrace(initial_attacker_observation_state=EmulationAttackerObservationState.schema(),
                             initial_defender_observation_state=EmulationDefenderObservationState.schema(),
                             emulation_name="")
        dto.attacker_observation_states = [EmulationAttackerObservationState.schema()]
        dto.defender_observation_states = [EmulationDefenderObservationState.schema()]
        dto.attacker_actions = [EmulationAttackerAction.schema()]
        dto.defender_actions = [EmulationDefenderAction.schema()]
        return dto

    def to_csv_record(self, max_time_steps: int, max_nodes: int, max_ports: int, max_vulns: int,
                      null_value: int = -1) -> Tuple[List[Union[str, int, float]], List[str]]:
        """
        Converts the trace into a csv row

        :param max_time_steps: the maximum number of time-steps to include in the row
        :param max_nodes: the maximum number of nodes to include metrics from
        :param max_ports: the maximum number of ports to include metrics from
        :param max_vulns: the maximum number of vulnerabilities to include metrics from
        :param null_value: the default null value if a metric is missing
        :return: the list of labels and values of the csv row
        """
        # lookup tables for vuln names, service names, protocol names
        labels = []
        values: List[Union[str, float]] = []
        intrusion_started = False
        attacker_observations = [self.initial_attacker_observation_state] + self.attacker_observation_states
        defender_observations = [self.initial_defender_observation_state] + self.defender_observation_states
        for t in range(max_time_steps):
            labels.append(f"{t}_intrusion")
            if len(attacker_observations) > t:
                if not intrusion_started and self.attacker_actions[t].id != EmulationAttackerActionId.CONTINUE:
                    intrusion_started = True
                values.append(int(intrusion_started))
            else:
                values.append(null_value)

            labels.append(f"{t}_attacker_action_id")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].id)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_name")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].name)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_type")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].type)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_ips")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append("-".join(self.attacker_actions[t - 1].ips))
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_index")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].index)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_outcome")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].action_outcome)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_action_execution_time")
            if (len(self.attacker_actions) + 1) > t:
                if t > 0:
                    values.append(self.attacker_actions[t - 1].execution_time)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_id")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].id)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_name")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].name)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_type")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].type)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_ips")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append("-".join(self.defender_actions[t - 1].ips))
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_index")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].index)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_outcome")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].action_outcome)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_action_execution_time")
            if (len(self.defender_actions) + 1) > t:
                if t > 0:
                    values.append(self.defender_actions[t - 1].execution_time)
                else:
                    values.append(null_value)
            else:
                values.append(null_value)
            labels.append(f"{t}_attacker_num_catched_flags")
            if len(attacker_observations) > t:
                values.append(attacker_observations[t].catched_flags)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_clients")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_client_population_metrics.num_clients)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_clients_arrival_rate")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_client_population_metrics.rate)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_pids")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.pids)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_cpu_percent")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.cpu_percent)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_mem_current")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.mem_current)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_mem_total")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.mem_total)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_mem_percent")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.mem_percent)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_blk_read")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.blk_read)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_blk_write")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.blk_write)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_net_rx")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.net_rx)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_net_tx")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_docker_stats.net_tx)
            else:
                values.append(null_value)
            for i in range(4):
                labels.append(f"{t}_defender_num_snort_priority_{i}_alerts")
                if len(defender_observations) > t:
                    values.append(defender_observations[t].avg_snort_ids_alert_counters.priority_alerts[i])
                else:
                    values.append(null_value)
            for i in range(len(set(collector_constants.SNORT_IDS_ROUTER.SNORT_ALERT_IDS_ID.values()))):
                labels.append(f"{t}_defender_num_snort_class_{i}_alerts")
                if len(defender_observations) > t:
                    values.append(defender_observations[t].avg_snort_ids_alert_counters.class_alerts[i])
                else:
                    values.append(null_value)
            labels.append(f"{t}_defender_num_snort_severe_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_snort_ids_alert_counters.severe_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_snort_warning_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_snort_ids_alert_counters.warning_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_snort_total_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_snort_ids_alert_counters.total_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_snort_alerts_weighted_by_priority")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_snort_ids_alert_counters.alerts_weighted_by_priority)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_ossec_severe_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_ossec_ids_alert_counters.severe_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_ossec_warning_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_ossec_ids_alert_counters.warning_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_ossec_total_alerts")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_ossec_ids_alert_counters.total_alerts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_ossec_alerts_weighted_by_level")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_ossec_ids_alert_counters.alerts_weighted_by_level)
            else:
                values.append(null_value)
            for i in range(16):
                labels.append(f"{t}_defender_num_ossec_level_{i}_alerts")
                if len(defender_observations) > t:
                    values.append(defender_observations[t].avg_ossec_ids_alert_counters.level_alerts[i])
                else:
                    values.append(null_value)
            for i in range(len(set(collector_constants.OSSEC.OSSEC_IDS_ALERT_GROUP_ID.values()))):
                labels.append(f"{t}_defender_num_ossec_group_{i}_alerts")
                if len(defender_observations) > t:
                    values.append(defender_observations[t].avg_ossec_ids_alert_counters.group_alerts[i])
                else:
                    values.append(null_value)
            labels.append(f"{t}_defender_num_logged_in_users")
            if len(defender_observations) > t:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_logged_in_users)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_failed_login_attempts")
            if len(defender_observations) > t and defender_observations[t].avg_aggregated_host_metrics is not None:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_failed_login_attempts)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_open_connections")
            if len(defender_observations) > t and defender_observations[t].avg_aggregated_host_metrics is not None:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_open_connections)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_login_events")
            if len(defender_observations) > t and defender_observations[t].avg_aggregated_host_metrics is not None:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_login_events)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_processes")
            if len(defender_observations) > t and defender_observations[t].avg_aggregated_host_metrics is not None:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_processes)
            else:
                values.append(null_value)
            labels.append(f"{t}_defender_num_users")
            if len(defender_observations) > t and defender_observations[t].avg_aggregated_host_metrics is not None:
                values.append(defender_observations[t].avg_aggregated_host_metrics.num_users)
            else:
                values.append(null_value)

            for i in range(max_nodes):
                labels.append(f"{t}_node_{i}_ip")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].ips[0])
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_os")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].os)
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_discovered")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(1)
                else:
                    values.append(null_value)
                for j in range(max_ports):
                    labels.append(f"{t}_attacker_node_{i}_port_{j}_port_number")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].ports) > j:
                        values.append(int(attacker_observations[t].machines[i].ports[j].port))
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_port_{j}_open")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].ports) > j:
                        values.append(int(attacker_observations[t].machines[i].ports[j].open))
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_port_{j}_service")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].ports) > j:
                        values.append(int(attacker_observations[t].machines[i].ports[j].service))
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_port_{j}_protocol")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].ports) > j:
                        values.append(int(attacker_observations[t].machines[i].ports[j].protocol))
                    else:
                        values.append(null_value)
                for k in range(max_vulns):
                    labels.append(f"{t}_attacker_node_{i}_vuln_{k}_name")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].cve_vulns) > k:
                        values.append(attacker_observations[t].machines[i].cve_vulns[k].name)
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_vuln_{k}_port_number")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].cve_vulns) > k:
                        values.append(int(attacker_observations[t].machines[i].cve_vulns[k].port))
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_vuln_{k}_protocol")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].cve_vulns) > k:
                        values.append(int(attacker_observations[t].machines[i].cve_vulns[k].protocol))
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_vuln_{k}_cvss")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].cve_vulns) > k:
                        values.append(attacker_observations[t].machines[i].cve_vulns[k].cvss)
                    else:
                        values.append(null_value)
                    labels.append(f"{t}_attacker_node_{i}_vuln_{k}_osvdbid")
                    if len(attacker_observations) > t and len(attacker_observations[t].machines) > i \
                            and len(attacker_observations[t].machines[i].cve_vulns) > k \
                            and attacker_observations[t].machines[i].cve_vulns[k].osvdb_id is not None:
                        values.append(int(attacker_observations[t].machines[i].cve_vulns[k].osvdb_id))
                    else:
                        values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_shell_access")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].shell_access))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_logged_in")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].logged_in))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_root")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].root))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_num_flags_found")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(len(attacker_observations[t].machines[i].flags_found))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_file_system_searched")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].filesystem_searched))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_untried_credentials")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].untried_credentials))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_telnet_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].telnet_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_ssh_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].ssh_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_ftp_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].ftp_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cassandra_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cassandra_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_irc_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].irc_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_mongo_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].mongo_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_mysql_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].mysql_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_smtp_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].smtp_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_postgres_brute_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].postgres_brute_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_tools_installed")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].tools_installed))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_backdoor_installed")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].backdoor_installed))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_backdoor_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].backdoor_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_dvwa_sql_injection_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].dvwa_sql_injection_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cve_2015_3306_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cve_2015_3306_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cve_2015_1427_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cve_2015_1427_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cve_2016_10033_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cve_2016_10033_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cve_2010_0426_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cve_2010_0426_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_attacker_node_{i}_cve_2015_5602_tried")
                if len(attacker_observations) > t and len(attacker_observations[t].machines) > i:
                    values.append(int(attacker_observations[t].machines[i].cve_2015_5602_tried))
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_pids")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.pids)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_cpu_percent")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.cpu_percent)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_mem_current")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.mem_current)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_mem_total")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.mem_total)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_mem_percent")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.mem_percent)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_blk_read")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.blk_read)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_blk_write")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.blk_write)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_net_rx")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.net_rx)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_net_tx")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].docker_stats.net_tx)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_logged_in_users")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_logged_in_users)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_failed_login_attempts")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_failed_login_attempts)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_open_connections")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_open_connections)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_login_events")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_login_events)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_processes")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_processes)
                else:
                    values.append(null_value)
                labels.append(f"{t}_node_{i}_defender_num_users")
                if len(defender_observations) > t and len(defender_observations[t].machines) > i:
                    values.append(defender_observations[t].machines[i].host_metrics.num_users)
                else:
                    values.append(null_value)
        assert len(values) == len(labels)
        return values, labels



from typing import List, Tuple, Dict, Any
import csle_common.constants.constants as constants
from csle_common.dao.emulation_config.container_network import ContainerNetwork
from csle_common.util.general_util import GeneralUtil
from csle_base.json_serializable import JSONSerializable


class NodeContainerConfig(JSONSerializable):
    """
    A DTO object representing an individual container in an emulation environment
    """

    def __init__(self, name: str, ips_and_networks: List[Tuple[str, ContainerNetwork]],
                 version: str, level: str, restart_policy: str,
                 suffix: str, os: str, execution_ip_first_octet: int = -1, docker_gw_bridge_ip: str = "",
                 physical_host_ip: str = ""):
        """
        Initializes the DTO

        :param name: the name of the node container
        :param ips_and_networks: the list of ips and networks that the container is connected to
        :param version: the version of the container
        :param level: the level of the container
        :param restart_policy: the restart policy of the container
        :param suffix: the suffix of the container id
        :param os: the operating system of the container
        :param execution_ip_first_octet: the first octet in the IP address (depends on the execution)
        :param docker_gw_bridge_ip: IP to reach the container from the host network
        :param physical_host_ip: IP of the physical host where the container is running
        """
        self.name = name
        self.ips_and_networks = ips_and_networks
        self.version = version
        self.level = level
        self.restart_policy = restart_policy
        self.suffix = suffix
        self.os = os
        self.execution_ip_first_octet = execution_ip_first_octet
        self.full_name_str = self.get_full_name()
        self.docker_gw_bridge_ip = docker_gw_bridge_ip
        self.physical_host_ip = physical_host_ip

    def get_ips(self) -> List[str]:
        """
        :return: a list of ips that this container has
        """
        return list(filter(lambda x: x is not None, map(lambda x: x[0], self.ips_and_networks)))

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "NodeContainerConfig":
        """
        Converts a dict representation to an instance
        :param d:  the dict to convert
        :return: the created instance
        """
        obj = NodeContainerConfig(
            name=d["name"],
            ips_and_networks=list(map(lambda x: (x[0], ContainerNetwork.from_dict(x[1])), d["ips_and_networks"])),
            version=d["version"], level=d["level"],
            restart_policy=d["restart_policy"], suffix=d["suffix"], os=d["os"],
            execution_ip_first_octet=d["execution_ip_first_octet"],
            docker_gw_bridge_ip=d["docker_gw_bridge_ip"], physical_host_ip=d["physical_host_ip"]
        )
        return obj

    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        d: Dict[str, Any] = {}
        d["name"] = self.name
        d["ips_and_networks"] = list(map(lambda x: (x[0], x[1].to_dict()), self.ips_and_networks))
        d["version"] = self.version
        d["restart_policy"] = self.restart_policy
        d["suffix"] = self.suffix
        d["os"] = self.os
        d["level"] = self.level
        d["full_name_str"] = self.get_full_name()
        d["execution_ip_first_octet"] = self.execution_ip_first_octet
        d["docker_gw_bridge_ip"] = self.docker_gw_bridge_ip
        d["physical_host_ip"] = self.physical_host_ip
        return d

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"name{self.name}, ips and networks: {self.ips_and_networks}, version: {self.version}, " \
               f"level:{self.level}, restart_policy: {self.restart_policy}, " \
               f"suffix:{self.suffix}, os:{self.os}, full_name:{self.full_name_str}, " \
               f"execution_ip_first_octet: {self.execution_ip_first_octet}," \
               f"docker_gw_bridge_ip: {self.docker_gw_bridge_ip}, physical_host_ip: {self.physical_host_ip}"

    def reachable(self, reachable_ips: List[str]) -> bool:
        """
        Check if container is reachable given a list of reachable ips

        :param reachable_ips: the list of reachable ips
        :return: True if the container is reachable, false otherwise
        """
        for ip in self.get_ips():
            if ip in reachable_ips:
                return True
        return False

    def get_full_name(self) -> str:
        """
        :return: the full name
        """
        return f"{self.name}{self.suffix}-{constants.CSLE.LEVEL}{self.level}-" \
               f"{self.execution_ip_first_octet}"

    def get_readable_name(self) -> str:
        """
        :return: the readable name
        """
        return f"csle-{self.name}{self.suffix}-{constants.CSLE.LEVEL}{self.level}_" \
               f"{self.execution_ip_first_octet}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "NodeContainerConfig":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return NodeContainerConfig.from_dict(json.loads(json_str))

    def copy(self) -> "NodeContainerConfig":
        """
        :return: a copy of the DTO
        """
        return NodeContainerConfig.from_dict(self.to_dict())

    def create_execution_config(self, ip_first_octet: int, physical_servers: List[str]) -> "NodeContainerConfig":
        """
        Creates a new config for an execution

        :param ip_first_octet: the first octet of the IP of the new execution
        :param physical_servers: the list of physical servers of the execution
        :return: the new config
        """
        config = self.copy()
        config.execution_ip_first_octet = ip_first_octet
        config.physical_host_ip = physical_servers[0]  # TODO Update this
        config.ips_and_networks = list(map(lambda x: (
            GeneralUtil.replace_first_octet_of_ip(ip=x[0], ip_first_octet=ip_first_octet),
            x[1].create_execution_config(ip_first_octet=ip_first_octet)), config.ips_and_networks))
        return config

    @staticmethod
    def schema() -> "NodeContainerConfig":
        """
        :return: get the schema of the DTO
        """
        return NodeContainerConfig(name="", ips_and_networks=[("", ContainerNetwork.schema())], version="", level="",
                                   restart_policy="", suffix="", os="", execution_ip_first_octet=-1)



from typing import Dict, Any
from abc import abstractmethod
from csle_common.dao.system_identification.system_model_type import SystemModelType
from csle_base.json_serializable import JSONSerializable


class SystemModel(JSONSerializable):
    """
    Abstract system model
    """

    def __init__(self, descr: str, model_type: SystemModelType):
        """
        Abstract constructor

        :param descr: the description of the system model
        :param model_type: the type of the system model
        """
        self.descr = descr
        self.model_type = model_type

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """
        Converts the object to a dict representation

        :return: a dict representation of the object
        """
        pass

    @staticmethod
    @abstractmethod
    def from_dict(d: Dict[str, Any]) -> "SystemModel":
        """
        Converts a dict representation of the object to an instance
        """
        pass

    @abstractmethod
    def copy(self) -> "SystemModel":
        """
        :return: a copy of the object
        """
        pass

    @staticmethod
    @abstractmethod
    def from_json_file(json_file_path: str) -> "SystemModel":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        pass



from typing import List, Dict, Union, Tuple, Any
import numpy as np
from numpy.typing import NDArray
import torch
import math
from torch.distributions import Categorical
from csle_common.models.fnn_w_softmax import FNNwithSoftmax
import csle_common.constants.constants as constants
from csle_common.dao.training.policy import Policy
from csle_common.dao.training.agent_type import AgentType
from csle_common.dao.training.player_type import PlayerType
from csle_common.dao.simulation_config.state import State
from csle_common.dao.simulation_config.state_type import StateType
from csle_common.dao.simulation_config.action import Action
from csle_common.dao.training.experiment_config import ExperimentConfig
from csle_common.dao.training.policy_type import PolicyType
from csle_common.logging.log import Logger


class FNNWithSoftmaxPolicy(Policy):
    """
    A feed-forward neural network policy with softmax output
    """

    def __init__(self, policy_network, simulation_name: str, save_path: str,
                 player_type: PlayerType, states: List[State],
                 actions: List[Action], experiment_config: ExperimentConfig, avg_R: float, input_dim: int,
                 output_dim: int):
        """
        Initializes the policy

        :param policy_network: the PPO model
        :param simulation_name: the simulation name
        :param save_path: the path to save the model to
        :param states: list of states (required for computing stage policies)
        :param actions: list of actions
        :param experiment_config: the experiment configuration for training the policy
        :param avg_R: the average reward of the policy when evaluated in the simulation
        """
        super(FNNWithSoftmaxPolicy, self).__init__(agent_type=AgentType.PPO, player_type=player_type)
        self.policy_network = policy_network
        self.id = -1
        self.simulation_name = simulation_name
        self.save_path = save_path
        self.states = states
        self.actions = actions
        self.experiment_config = experiment_config
        self.avg_R = avg_R
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.policy_type = PolicyType.FNN_W_SOFTMAX
        if self.policy_network is None and self.save_path != "":
            try:
                hidden_dim = \
                    self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_NEURONS_PER_HIDDEN_LAYER].value
                hidden_layers = self.experiment_config.hparams[constants.NEURAL_NETWORKS.NUM_HIDDEN_LAYERS].value
                activation = self.experiment_config.hparams[constants.NEURAL_NETWORKS.ACTIVATION_FUNCTION].value
                if not isinstance(hidden_dim, int):
                    raise ValueError(f"The number of neurons must be an integer, got type: {type(hidden_dim)}")
                if not isinstance(hidden_layers, int):
                    raise ValueError(f"The number of layers must be an integer, got type: {type(hidden_layers)}")
                if not isinstance(activation, str):
                    raise ValueError(f"The type of the actvation function must be a string, "
                                     f"got type: {type(activation)}")
                self.policy_network = FNNwithSoftmax(
                    input_dim=input_dim,
                    output_dim=output_dim,
                    hidden_dim=hidden_dim,
                    num_hidden_layers=hidden_layers,
                    hidden_activation=activation)
                self.policy_network = self.policy_network.load_state_dict(torch.load(self.save_path))
            except Exception as e:
                Logger.__call__().get_logger().warning(
                    f"There was an exception loading the model from path: {self.save_path}, "
                    f"exception: {str(e)}, {repr(e)}")

    def action(self, o: NDArray[Any], deterministic: bool = True) -> Any:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :param deterministic: boolean flag indicating whether the action selection should be deterministic
        :return: the selected action
        """
        state = torch.from_numpy(o).float()
        device = self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value
        if not isinstance(device, str):
            raise ValueError(f"Device must be a str, got type: {type(device)}")
        state1 = state.to(str(torch.device(device)))

        # Forward pass using the current policy network to predict P(a|s)
        action_probs = self.policy_network(state1).squeeze()
        # Set probability of non-legal actions to 0
        action_probs_1 = action_probs.clone()

        # Use torch.distributions package to create a parameterizable probability distribution of the learned policy
        # PG uses a trick to turn the gradient into a stochastic gradient which we can sample from in order to
        # approximate the true gradient (which we can’t compute directly). It can be seen as an alternative to the
        # reparameterization trick
        policy_dist = Categorical(action_probs_1)

        # Sample an action from the probability distribution
        action = policy_dist.sample()
        return action

    def get_action_and_log_prob(self, state: NDArray[Any]) -> Tuple[int, float]:
        """
        Samples an action from the policy network

        :param policy_network: the policy network
        :param state: the state to sample an action for
        :return: The sampled action id and the log probability
        """
        state1 = torch.from_numpy(state.flatten()).float()
        device = self.experiment_config.hparams[constants.NEURAL_NETWORKS.DEVICE].value
        if not isinstance(device, str):
            raise ValueError(f"Device must be a str, got type: {type(device)}")
        state2 = state1.to(torch.device(device))

        # Forward pass using the current policy network to predict P(a|s)
        action_probs = self.policy_network(state2).squeeze()
        # Set probability of non-legal actions to 0
        action_probs_1 = action_probs.clone()

        # Use torch.distributions package to create a parameterizable probability distribution of the learned policy
        # PG uses a trick to turn the gradient into a stochastic gradient which we can sample from in order to
        # approximate the true gradient (which we can’t compute directly). It can be seen as an alternative to the
        # reparameterization trick
        policy_dist = Categorical(action_probs_1)

        # Sample an action from the probability distribution
        action = policy_dist.sample()

        # log_prob returns the log of the probability density/mass function evaluated at value.
        # save the log_prob as it will use later on for computing the policy gradient
        # policy gradient theorem says that the stochastic gradient of the expected return of the current policy is
        # the log gradient of the policy times the expected return, therefore we save the log of the policy distribution
        # now and use it later to compute the gradient once the episode has finished.
        log_prob = policy_dist.log_prob(action)
        return action.item(), log_prob

    def probability(self, o: NDArray[Any], a: int) -> Union[int, float]:
        """
        Multi-threshold stopping policy

        :param o: the current observation
        :return: the selected action
        """
        action, log_prob = self.get_action_and_log_prob(state=o)
        if action == a:
            return math.exp(log_prob)
        else:
            return 0

    def to_dict(self) -> Dict[str, Any]:
        """
        :return: a dict representation of the policy
        """
        d: Dict[str, Any] = {}
        d["id"] = self.id
        d["simulation_name"] = self.simulation_name
        d["save_path"] = self.save_path
        d["states"] = list(map(lambda x: x.to_dict(), self.states))
        d["player_type"] = self.player_type
        d["actions"] = list(map(lambda x: x.to_dict(), self.actions))
        d["experiment_config"] = self.experiment_config.to_dict()
        d["agent_type"] = self.agent_type
        d["avg_R"] = self.avg_R
        d["input_dim"] = self.input_dim
        d["output_dim"] = self.output_dim
        d["policy_type"] = self.policy_type
        return d

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "FNNWithSoftmaxPolicy":
        """
        Converst a dict representation of the object to an instance

        :param d: the dict to convert
        :return: the created instance
        """
        obj = FNNWithSoftmaxPolicy(policy_network=None, simulation_name=d["simulation_name"],
                                   save_path=d["save_path"],
                                   states=list(map(lambda x: State.from_dict(x), d["states"])),
                                   player_type=d["player_type"],
                                   actions=list(map(lambda x: Action.from_dict(x), d["actions"])),
                                   experiment_config=ExperimentConfig.from_dict(d["experiment_config"]),
                                   avg_R=d["avg_R"], input_dim=d["input_dim"], output_dim=d["output_dim"])
        # policy_type=d["policy_type"])
        obj.id = d["id"]
        return obj

    def stage_policy(self, o: Union[List[int], List[float]]) -> List[List[float]]:
        """
        Gets the stage policy, i.e a |S|x|A| policy

        :param o: the latest observation
        :return: the |S|x|A| stage policy
        """
        b1 = o[1]
        l = int(o[0])
        if not self.player_type == PlayerType.ATTACKER:
            stage_policy = []
            for _ in self.states:
                stage_policy.append(self._get_attacker_dist(obs=o))
            return stage_policy
        else:
            stage_policy = []
            for s in self.states:
                if s.state_type != StateType.TERMINAL:
                    o = [l, b1, s.id]
                    stage_policy.append(self._get_attacker_dist(obs=o))
                else:
                    stage_policy.append([0.5, 0.5])
            return stage_policy

    def _get_attacker_dist(self, obs) -> List[float]:
        """
        Utility function for getting the action distribution conditioned on a given observation

        :param obs: the observation to condition on
        :return: the conditional ation distribution
        """
        obs = np.array([obs])
        actions, values, log_prob = self.policy_network.policy.forward(
            obs=torch.tensor(obs).to(self.policy_network.device))
        action = actions[0]
        if action == 1:
            stop_prob = math.exp(log_prob)
        else:
            stop_prob = 1 - math.exp(log_prob)
        return [1 - stop_prob, stop_prob]

    def __str__(self) -> str:
        """
        :return: a string representation of the object
        """
        return f"policy network: {self.policy_network}, id: {self.id}, simulation_name: {self.simulation_name}, " \
               f"save path: {self.save_path}, states: {self.states}, experiment_config: {self.experiment_config}," \
               f"avg_R: {self.avg_R}, input_dim: {self.input_dim}, output_dim: {self.output_dim}"

    @staticmethod
    def from_json_file(json_file_path: str) -> "FNNWithSoftmaxPolicy":
        """
        Reads a json file and converts it to a DTO

        :param json_file_path: the json file path
        :return: the converted DTO
        """
        import io
        import json
        with io.open(json_file_path, 'r') as f:
            json_str = f.read()
        return FNNWithSoftmaxPolicy.from_dict(json.loads(json_str))

    def save_policy_network(self) -> None:
        """
        Saves the PyTorch Model Weights

        :return: None
        """
        path = self.save_path
        torch.save(self.policy_network.state_dict(), path)

    def copy(self) -> "FNNWithSoftmaxPolicy":
        """
        :return: a copy of the DTO
        """
        return self.from_dict(self.to_dict())
